from math import cos, sin, radians

from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.AllChem import Compute2DCoords

from luna.wrappers.base import MolWrapper
from luna.util.default_values import ATOM_TYPES_COLOR


class PharmacophoreDepiction:
    """Draw molecules and depict pharmacophoric properties as colored circles.

    Parameters
    ----------
    feature_extractor : :class:`~luna.mol.features.FeatureExtractor`
        Perceive pharmacophoric properties from molecules.
    colors : :class:`~luna.util.ColorPallete`
        Color scheme for pharmacophoric properties perceived by ``feature_extractor``.
        The default value is :const:`~luna.util.default_values.ATOM_TYPES_COLOR`.
    format : {'png', 'svg'}
        The output file format. The default value is 'png'.
    figsize : tuple of (float, float)
        Width and height in inches. The default value is (800, 800).
    font_size : float
        The font size. The units are, roughly, pixels. The default value is 0.5.
    circle_dist : float
        Distance between circles (pharmacophoric properties). The default value is 0.2.
    circle_radius :
        Circles' radius size of pharmacophoric properties. The default value is 0.3.
    use_bw_atom_palette : bool
        Use a black & white palette for atoms and bonds.

    Examples
    --------

    First, let's read a molecule (glutamine).

    >>> from luna.wrappers.base import MolWrapper
    >>> mol = MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O")

    Now, create a feature factory and instantiate a new FeatureExtractor object.

    >>> from luna.util.default_values import ATOM_PROP_FILE
    >>> from rdkit.Chem import ChemicalFeatures
    >>> from luna.mol.features import FeatureExtractor
    >>> feature_factory = ChemicalFeatures.BuildFeatureFactory(ATOM_PROP_FILE)
    >>> feature_extractor = FeatureExtractor(feature_factory)

    Instantiate a new PharmacophoreDepiction object with the desired configuration.
    For example, you can provide a color scheme for pharmacophoric properties, the image size, and its format.

    >>> from luna.util.default_values import ATOM_TYPES_COLOR
    >>> from luna.mol.depiction import PharmacophoreDepiction
    pd = PharmacophoreDepiction(feature_extractor=feature_extractor, colors=ATOM_TYPES_COLOR,
                                fig_size=(500, 500), format="svg")

    Finally, you can draw the molecule with annotated pharmacophoric properties.

    >>> pd.plot_fig(mol, "output.svg")
    """

    def __init__(self, feature_extractor=None, colors=ATOM_TYPES_COLOR, format="png",
                 fig_size=(800, 800), font_size=0.5, circle_dist=0.2, circle_radius=0.3,
                 use_bw_atom_palette=True):

        self.feature_extractor = feature_extractor
        self.colors = colors
        self.format = format
        self.fig_size = fig_size
        self.font_size = font_size
        self.circle_dist = circle_dist
        self.circle_radius = circle_radius
        self.use_bw_atom_palette = use_bw_atom_palette

    def _perceive_atm_types(self, rdmol):
        if self.feature_extractor is not None:
            return self.feature_extractor.get_features_by_atoms(rdmol)
        return {}

    def plot_fig(self, mol_obj, output=None, atm_types=None, legend=None):
        """Draw the molecule ``mol_obj`` and depict its pharmacophoric properties.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
            The molecule.
        output : str
            The output file where the molecule will be drawn.
            If None, returns a drawing object (:class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDraw2DCairo` or
            :class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDraw2DSVG`).
        atm_types : dict or None
            A pre-annotated dictionary for mapping atoms and pharmacophoric properties.
            If None, try to perceive properties with ``feature_extractor``.
        legend : str
            A title for the figure.

        Returns
        -------
        drawer : None or a drawing object (:class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDraw2DCairo` or \
        :class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDraw2DSVG`)
        """

        rdmol = MolWrapper(mol_obj).as_rdkit()

        # Make a copy of the molecule
        rwm = Chem.RWMol(rdmol)
        Compute2DCoords(rwm)

        if atm_types is None:
            atm_types = self._perceive_atm_types(rdmol)

        if self.format == "png":
            drawer = rdMolDraw2D.MolDraw2DCairo(*self.fig_size)
        elif self.format == "svg":
            drawer = rdMolDraw2D.MolDraw2DSVG(*self.fig_size)
        else:
            # raise
            pass

        opts = drawer.drawOptions()

        highlight = {}
        for atm_id in atm_types:
            centroid = list(rwm.GetConformer().GetAtomPosition(atm_id))
            valid_features = [f for f in atm_types[atm_id] if f.name in self.colors]

            if valid_features:
                if len(valid_features) == 1:
                    pos = centroid
                    atmIdx = self._add_dummy_atom(rwm, centroid)
                    highlight[atmIdx] = self.colors.get_normalized_color(valid_features[0].name)
                    opts.atomLabels[atmIdx] = ''
                else:
                    sliceRad = radians(360 / len(valid_features))
                    for i, feature in enumerate(valid_features):
                        rad = i * sliceRad
                        pos = [self.circle_dist * cos(rad), self.circle_dist * sin(rad), 0]
                        adj_Pos = [x + y for x, y in zip(centroid, pos)]
                        new_atm_id = self._add_dummy_atom(rwm, adj_Pos)
                        highlight[new_atm_id] = self.colors.get_normalized_color(feature.name)
                        opts.atomLabels[new_atm_id] = ''

        atoms = [x for x in highlight]
        radius = {a: self.circle_radius for a in atoms}

        opts.flagCloseContactsDist = -1000
        opts.legendFontSize = 20
        opts.padding = 0.02

        if self.use_bw_atom_palette:
            opts.useBWAtomPalette()
        else:
            opts.useDefaultAtomPalette()

        legend = legend or ""

        drawer.SetFontSize(self.font_size)
        drawer.DrawMolecule(rwm, highlightAtoms=atoms, highlightAtomColors=highlight,
                            highlightBonds=[], highlightAtomRadii=radius, legend=legend)
        drawer.FinishDrawing()

        if output:
            if self.format == "png":
                drawer.WriteDrawingText(output)
            elif self.format == "svg":
                svg = drawer.GetDrawingText().replace('svg:', '')
                with open(output, "w") as fh:
                    fh.write(svg)
        else:
            return drawer

    def _add_dummy_atom(self, mol, pos=None):
        new_atm = Chem.rdchem.Atom(10)
        atm_id = mol.AddAtom(new_atm)
        mol.GetConformer().SetAtomPosition(atm_id, pos)
        new_atm.SetNoImplicit(True)

        return atm_id



from luna.wrappers.base import AtomWrapper

import logging
logger = logging.getLogger()


# TODO: implement this other model: https://github.com/openbabel/openbabel/blob/master/src/formats/mdlvalence.h


class ChargeModel:
    """Implementation of a charge model."""

    def get_charge(self):
        raise NotImplementedError("Subclasses should implement this.")


class OpenEyeModel(ChargeModel):
    """Implementation of OpenEye charge model."""

    def get_charge(self, atm_obj):
        """Get the formal charge for atom ``atom_obj``.

        Currently, only formal charges for the elements Hydrogen, Carbon, Nitrogen, Oxygen, Phosphorus,
        Sulfur, Chlorine, Fluorine, Bromine, Iodine, Magnesium, Calcium, Zinc, Lithium, Sodium,
        Potassium, and Boron can be recovered.

        Parameters
        ----------
        atm_obj : :class:`~luna.wrappers.base.AtomWrapper`
            The target atom.

        Examples
        --------
        >>> from luna.mol.charge_model import OpenEyeModel
        >>> from luna.wrappers.base import MolWrapper
        >>> cm = OpenEyeModel()
        >>> mol_obj = MolWrapper.from_smiles("[NH3+]CC([O-])=O")
        >>> for atm_obj in mol_obj.get_atoms():
        >>>     formal_charge = cm.get_charge(atm_obj)
        >>>     print("Charge for atom #%d (%s): %d" % (atm_obj.get_idx(),
        ...                                             atm_obj.get_symbol(),
        ...                                             formal_charge))
        Charge for atom #0 (N): 1.
        Charge for atom #1 (C): 0.
        Charge for atom #2 (C): 0.
        Charge for atom #3 (O): -1.
        Charge for atom #4 (O): 0.

        """
        formal_charge = None

        atm_num = atm_obj.get_atomic_num()
        valence = atm_obj.get_valence()

        # Hydrogen
        if atm_num == 1:
            if valence != 1:
                formal_charge = 1
            elif valence == 1:
                formal_charge = 0
        # Carbon
        elif atm_num == 6:
            if valence == 3:
                has_polar_nb = False
                nbs = atm_obj.get_neighbors()
                for nb_atm_obj in nbs:
                    nb_atm_obj = AtomWrapper(nb_atm_obj)
                    # Polar neighbor: N, O or S
                    if nb_atm_obj.get_atomic_num() in (7, 8, 16):
                        has_polar_nb = True
                        break
                if has_polar_nb is True:
                    formal_charge = 1
                else:
                    formal_charge = -1
            elif valence == 4:
                formal_charge = 0
        # Nitrogen
        elif atm_num == 7:
            if valence == 2:
                formal_charge = -1
            elif valence == 4:
                formal_charge = 1
            elif valence == 3:
                formal_charge = 0
        # Oxygen
        elif atm_num == 8:
            if valence == 1:
                formal_charge = -1
            elif valence == 3:
                formal_charge = 1
            elif valence == 2:
                formal_charge = 0
        # Phosphorus
        elif atm_num == 15:
            if valence == 4:
                formal_charge = 1
        # Sulfur
        elif atm_num == 16:
            if valence == 1:
                formal_charge = -1
            elif valence == 3:
                formal_charge = 1
            elif valence == 5:
                formal_charge = -1
            elif valence == 4:
                if atm_obj.get_degree() == 4:
                    formal_charge = 2
            elif valence == 2 or valence == 6:
                formal_charge == 0
        # Chlorine
        elif atm_num == 17:
            if valence == 0:
                formal_charge = -1
            elif valence == 4:
                formal_charge = 3
            elif valence == 1:
                formal_charge = 0
        # Fluorine, Bromine, Iodine
        elif atm_num == 9 or atm_num == 35 or atm_num == 53:
            if valence == 0:
                formal_charge = -1
            elif valence == 1:
                formal_charge = 0
        # Magnesium, Calcium, Zinc
        elif atm_num == 12 or atm_num == 20 or atm_num == 30:
            if valence == 0:
                formal_charge = 2
            elif valence == 2:
                formal_charge = 0
        # Lithium, Sodium, Potassium
        elif atm_num == 3 or atm_num == 11 or atm_num == 19:
            if valence == 0:
                formal_charge = 1
            elif valence == 1:
                formal_charge = 0
        # Boron
        elif atm_num == 5:
            # If the valence is four, the formal charge is -1.
            # OBS: there is an error in OpenEye chargel model text, they said that Boron should have a
            # charge of +1 when the valence is 4. However, the MDL Valence Model says it should be -1.
            if valence == 4:
                formal_charge = -1
            elif valence == 3:
                formal_charge = 0
        # Iron: 26
        # If the valence is zero, the formal charge is +3 if the partial charge is 3.0, and +2 otherwise.
        # Copper: 29
        # If the valence is zero, the formal charge is +2 if the partial charge is 2.0, and +1 otherwise.
        # else:
        # For the remaining elements, if the valence of an atom is zero, its formal charge is set from its partial charge.

        return formal_charge



from os import path

from luna.interaction.config import DefaultInteractionConfig, InteractionConfig
from luna.util import ColorPallete


ENTRY_SEPARATOR = ":"

LUNA_PATH = path.abspath(path.join(path.realpath(__file__), '../../../'))
OUTPUT_PATH = "%s/output/public" % LUNA_PATH
PDB_PATH = "%s/pdb" % OUTPUT_PATH
TMP_FILES = "%s/tmp" % OUTPUT_PATH

CONFIG_PATH = path.abspath(path.join(path.realpath(__file__), '../../', 'data'))
DB_CONFIG_FILE = "%s/mysql.ini" % CONFIG_PATH
ATOM_PROP_FILE = "%s/LUNA.fdef" % CONFIG_PATH
LIGAND_EXPO_FILE = "%s/ligand_expo.tsv" % CONFIG_PATH
MIN_FDEF_FILE = '%s/MinimalFeatures.fdef' % CONFIG_PATH

INTERACTION_CONFIG = DefaultInteractionConfig()
BOUNDARY_CONFIG = InteractionConfig({"bsite_cutoff": 6.2,
                                     "cache_cutoff": 10})

NMR_METHODS = ["SOLID-STATE NMR", "SOLUTION NMR"]

COV_SEARCH_RADIUS = 2.2

IFP_LENGTH = 4096

ACCEPTED_MOL_OBJ_TYPES = ("rdkit", "openbabel")

OPENBABEL = "obabel"

RECURSION_LIMIT = 600000

ATOM_TYPES_COLOR = ColorPallete({
    "Aromatic": (255, 187, 204),
    "Acceptor": (158, 154, 200),
    "Donor": (17, 170, 119),
    "Hydrophobe": (254, 224, 144),
    "Hydrophobic": (254, 224, 144),
    "Negative": (251, 128, 114),
    "Positive": (31, 120, 180),
    "NegIonizable": (251, 128, 114),
    "PosIonizable": (31, 120, 180),
    "NegativelyIonizable": (251, 128, 114),
    "PositivelyIonizable": (31, 120, 180),
    "HalogenDonor": (196, 156, 148),
    "Metal": (96, 125, 139),
    "WeakDonor": (161, 217, 155),
    "WeakAcceptor": (188, 194, 220),
    "Electrophile": (251, 138, 124),
    "Nucleophile": (158, 218, 229),
    "ChalcogenDonor": (255, 193, 7),
}, (255, 255, 255))

CHEMICAL_FEATURE_IDS = {
    "Aromatic": 1,
    "Acceptor": 2,
    "Donor": 3,
    "Hydrophobe": 4,
    "Hydrophobic": 5,
    "Negative": 6,
    "Positive": 7,
    "Negatively ionizable": 8,
    "Positively ionizable": 9,
    "Halogen donor": 10,
    "Metal": 12,
    "Lumped hydrophobe": 13,
    "Weak donor": 14,
    "Weak acceptor": 15,
    "Electrophile": 16,
    "Nucleophile": 17,
    "Chalcogen donor": 19,
    "Amide": 21,
    "Atom": 22
}

INTERACTION_IDS = {
    "Proximal": 0,
    "Hydrogen bond": 1,
    "Ionic": 2,
    "Salt bridge": 3,
    "Cation-pi": 4,
    "Hydrophobic": 5,
    "Halogen bond": 6,
    "Repulsive": 7,
    "Water-bridged hydrogen bond": 8,
    "Amide-aromatic stacking": 9,
    "Weak hydrogen bond": 10,
    "Covalent bond": 11,
    "Atom overlap": 12,
    "Van der Waals clash": 13,
    "Van der Waals": 14,
    "Chalcogen bond": 15,
    "Chalcogen-pi": 16,
    "Halogen-pi": 17,
    "Orthogonal multipolar": 18,
    "Parallel multipolar": 19,
    "Antiparallel multipolar": 20,
    "Tilted multipolar": 21,
    "Multipolar": 22,
    "Cation-nucleophile": 23,
    "Anion-electrophile": 24,
    "Unfavorable anion-nucleophile": 25,
    "Unfavorable cation-electrophile": 26,
    "Unfavorable nucleophile-nucleophile": 27,
    "Unfavorable electrophile-electrophile": 28,
    "Pi-stacking": 29,
    "Face-to-face pi-stacking": 30,
    "Face-to-edge pi-stacking": 31,
    "Face-to-slope pi-stacking": 32,
    "Edge-to-edge pi-stacking": 33,
    "Edge-to-face pi-stacking": 34,
    "Edge-to-slope pi-stacking": 35,
    "Displaced face-to-face pi-stacking": 36,
    "Displaced face-to-edge pi-stacking": 37,
    "Displaced face-to-slope pi-stacking": 38,
    "Single bond": 39,
    "Double bond": 40,
    "Triple bond": 41,
    "Aromatic bond": 42,
    "Other bond": 43
}

INTERACTION_SHORT_NAMES = {
    "Proximal": "prox",
    "Hydrogen bond": "hbond",
    "Ionic": "ionic",
    "Salt bridge": "salt_bridge",
    "Cation-pi": "cation-pi",
    "Hydrophobic": "hphobe",
    "Halogen bond": "xbond",
    "Halogen-pi": "x-pi",
    "Chalcogen bond": "ybond",
    "Chalcogen-pi": "y-pi",
    "Repulsive": "repuls",
    "Water-bridged hydrogen bond": "water_hbond",
    "Amide-aromatic stacking": "amide_stack",
    "Weak hydrogen bond": "weak_hbond",
    "Orthogonal multipolar": "ort_multipol",
    "Parallel multipolar": "par_multipol",
    "Antiparallel multipolar": "antipar_multipol",
    "Tilted multipolar": "tilted_multipol",
    "Multipolar": "multipol",
    "Unfavorable nucleophile-nucleophile": "unf_nucleop_nucleop",
    "Unfavorable electrophile-electrophile": "unf_electrop_electrop",
    "Cation-nucleophile": "cat_nucleop",
    "Anion-electrophile": "ani_electrop",
    "Unfavorable anion-nucleophile": "unf_ani_nucleop",
    "Unfavorable cation-electrophile": "unf_cat_electrop",
    "Covalent bond": "cov",
    "Atom overlap": "atm_overlap",
    "Van der Waals clash": "vdw_clash",
    "Van der Waals": "vdw",
    "Pi-stacking": "pi-stack",
    "Face-to-face pi-stacking": "face-to-face_stack",
    "Face-to-edge pi-stacking": "face-to-edge_stack",
    "Face-to-slope pi-stacking": "face-to-slope_stack",
    "Edge-to-edge pi-stacking": "edge-to-edge_stack",
    "Edge-to-face pi-stacking": "edge-to-face_stack",
    "Edge-to-slope pi-stacking": "edge-to-slope_stack",
    "Displaced face-to-face pi-stacking": "disp_face-to-face_stack",
    "Displaced face-to-edge pi-stacking": "disp_face-to-edge_stack",
    "Displaced face-to-slope pi-stacking": "disp_face-to-slope_stack",
    "Single bond": "single-bond",
    "Double bond": "double-bond",
    "Triple bond": "triple-bond",
    "Aromatic bond": "aromatic-bond",
    "Other bond": "other-bond",
}

PYMOL_INTERACTION_COLOR = ColorPallete({
    "Proximal": "gray60",
    "Hydrogen bond": "tv_blue",
    "Water-bridged hydrogen bond": "lightblue",
    "Weak hydrogen bond": "lightteal",
    "Ionic": "green",
    "Salt bridge": "forest",
    "Cation-pi": "salmon",
    "Amide-aromatic stacking": "raspberry",
    "Hydrophobic": "orange",
    "Halogen bond": "aquamarine",
    "Halogen-pi": "aquamarine",
    "Chalcogen bond": "lightorange",
    "Chalcogen-pi": "lightorange",
    "Repulsive": "violetpurple",
    "Covalent bond": "black",
    "Atom overlap": "gray40",
    "Van der Waals clash": "gray90",
    "Van der Waals": "gray50",
    "Orthogonal multipolar": "paleyellow",
    "Parallel multipolar": "paleyellow",
    "Antiparallel multipolar": "paleyellow",
    "Tilted multipolar": "paleyellow",
    "Multipolar": "paleyellow",
    "Unfavorable nucleophile-nucleophile": "dirtyviolet",
    "Unfavorable electrophile-electrophile": "dirtyviolet",
    "Cation-nucleophile": "palegreen",
    "Anion-electrophile": "palegreen",
    "Unfavorable anion-nucleophile": "dirtyviolet",
    "Unfavorable cation-electrophile": "dirtyviolet",
    "Pi-stacking": "tv_red",
    "Face-to-face pi-stacking": "tv_red",
    "Face-to-edge pi-stacking": "tv_red",
    "Face-to-slope pi-stacking": "tv_red",
    "Edge-to-edge pi-stacking": "tv_red",
    "Edge-to-face pi-stacking": "tv_red",
    "Edge-to-slope pi-stacking": "tv_red",
    "Displaced face-to-face pi-stacking": "tv_red",
    "Displaced face-to-edge pi-stacking": "tv_red",
    "Displaced face-to-slope pi-stacking": "tv_red",
    "Single bond": "black",
    "Double bond": "black",
    "Triple bond": "black",
    "Aromatic bond": "black",
    "Other bond": "black",
}, "white")

PYMOL_INTERACTION_COLOR_AS_RGB = ColorPallete({
    "Proximal": (0.6, 0.6, 0.6),
    "Hydrogen bond": (0.3, 0.3, 1.0),
    "Water-bridged hydrogen bond": (0.75, 0.75, 1.0),
    "Weak hydrogen bond": (0.4, 0.7, 0.7),
    "Ionic": (0.0, 1.0, 0.0),
    "Salt bridge": (0.2, 0.6, 0.2),
    "Cation-pi": (1.0, 0.6, 0.6),
    "Amide-aromatic stacking": (0.70, 0.30, 0.40),
    "Hydrophobic": (1.0, 0.5, 0.0),
    "Halogen bond": (0.5, 1, 1),
    "Halogen-pi": (0.5, 1, 1),
    "Chalcogen bond": (1.0, 0.8, 0.5),
    "Chalcogen-pi": (1.0, 0.8, 0.5),
    "Repulsive": (0.55, 0.25, 0.60),
    "Covalent bond": (0.0, 0.0, 0.0),
    "Atom overlap": (0.4, 0.4, 0.4),
    "Van der Waals clash": (0.9, 0.9, 0.9),
    "Van der Waals": (0.5, 0.5, 0.5),
    "Orthogonal multipolar": (1.0, 1.0, 0.5),
    "Parallel multipolar": (1.0, 1.0, 0.5),
    "Antiparallel multipolar": (1.0, 1.0, 0.5),
    "Tilted multipolar": (1.0, 1.0, 0.5),
    "Multipolar": (1.0, 1.0, 0.5),
    "Unfavorable nucleophile-nucleophile": (0.70, 0.50, 0.50),
    "Unfavorable electrophile-electrophile": (0.70, 0.50, 0.50),
    "Cation-nucleophile": (0.65, 0.9, 0.65),
    "Anion-electrophile": (0.65, 0.9, 0.65),
    "Unfavorable anion-nucleophile": (0.70, 0.50, 0.50),
    "Unfavorable cation-electrophile": (0.70, 0.50, 0.50),
    "Pi-stacking": (1.0, 0.2, 0.2),
    "Face-to-face pi-stacking": (1.0, 0.2, 0.2),
    "Face-to-edge pi-stacking": (1.0, 0.2, 0.2),
    "Face-to-slope pi-stacking": (1.0, 0.2, 0.2),
    "Edge-to-edge pi-stacking": (1.0, 0.2, 0.2),
    "Edge-to-face pi-stacking": (1.0, 0.2, 0.2),
    "Edge-to-slope pi-stacking": (1.0, 0.2, 0.2),
    "Displaced face-to-face pi-stacking": (1.0, 0.2, 0.2),
    "Displaced face-to-edge pi-stacking": (1.0, 0.2, 0.2),
    "Displaced face-to-slope pi-stacking": (1.0, 0.2, 0.2),
    "Single bond": (0.0, 0.0, 0.0),
    "Double bond": (0.0, 0.0, 0.0),
    "Triple bond": (0.0, 0.0, 0.0),
    "Aromatic bond": (0.0, 0.0, 0.0),
    "Other bond": (0.0, 0.0, 0.0),
}, (1.0, 1.0, 1.0))


ARTIFACTS_LIST = ["ACE", "ACT", "BME", "CSD", "CSW", "EDO", "FMT", "GOL", "MSE", "NAG", "NO3", "PO4", "SGM", "SO4", "TPO"]



import re

from luna.wrappers.base import MolWrapper
from luna.util.default_values import MIN_FDEF_FILE
from luna.util.exceptions import FingerprintNotCreated, IllegalArgumentError

from rdkit.Chem.Fingerprints import FingerprintMols
from rdkit.Chem import MACCSkeys
from rdkit.Chem.AtomPairs import Pairs, Torsions
from rdkit.Chem import AllChem
from rdkit.Chem import ChemicalFeatures
from rdkit.Chem.Pharm2D import Generate
from rdkit.Chem.Pharm2D.SigFactory import SigFactory

import logging

logger = logging.getLogger()


class FingerprintGenerator():
    """Generate molecular fingerprints for the molecule ``mol``.

    Parameters
    ----------
    mol : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`, optional
        The molecule.

    Examples
    --------

    First, create a new `FingerprintGenerator` object.

    >>> from luna.mol.fingerprint import FingerprintGenerator
    >>> fg = FingerprintGenerator()

    Now, let's read a molecule (glutamine) and set it to the `FingerprintGenerator` object.

    >>> from luna.wrappers.base import MolWrapper
    >>> fg.mol = MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O")

    Finally, you can call any available function to generate the desired fingerprint type.
    In the below example, a count ECFP4 fingerprint of size 1,024 is created.

    >>> fp = fg.morgan_fp(radius=2, length=1024, type=2)
    >>> print(fp.GetNonzeroElements())
    {1: 1, 80: 2, 140: 1, 147: 2, 389: 1, 403: 1, 540: 1, 545: 1, 650: 2, 728: 1, 739: 1, 767: 1, 786: 1, 807: 3, 820: 1, 825: 1, 874: 1, 893: 2, 900: 1}

    You can then continue using the `FingerprintGenerator` object to create other fingerprint types.
    For example, let's create a 2D pharmacophore fingerprint.

    >>> fp = fg.pharm2d_fp()
    >>> print(fp.GetNumOnBits())
    90
    """

    def __init__(self, mol_obj=None):
        self.mol = mol_obj

    @property
    def mol(self):
        """:class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`: The molecule."""
        return self._mol_obj

    @mol.setter
    def mol(self, mol_obj):
        self._mol_obj = mol_obj

        if mol_obj:
            self._rdmol = MolWrapper(mol_obj).as_rdkit()

    def rdk_fp(self):
        """Generate an RDKit topological fingerprint for the molecule ``mol``.

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        """
        try:
            return FingerprintMols.FingerprintMol(self._rdmol)
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("Fingerprint could not be created.")

    def maccs_keys_fp(self):
        """Generate a MACCS keys fingerprint for the molecule ``mol``.

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        """
        try:
            return MACCSkeys.GenMACCSKeys(self._rdmol)
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("Fingerprint could not be created.")

    def atom_pairs_fp(self):
        """Generate an atom pairs fingerprint for the molecule ``mol``.

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        """
        try:
            return Pairs.GetAtomPairFingerprint(self._rdmol)
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("Fingerprint could not be created.")

    def torsion_fp(self):
        """Generate a topological torsion fingerprint for the molecule ``mol``.

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        """
        try:
            return Torsions.GetTopologicalTorsionFingerprintAsIntVect(self._rdmol)
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("Fingerprint could not be created.")

    def morgan_fp(self, radius=2, length=2048, features=False, type=1):
        """Generate a Morgan fingerprint for the molecule ``mol``.

        Parameters
        ----------
        radius : int
            Define the maximum radius of the circular neighborhoods considered for each atom.
            The default value is 2, which is roughly equivalent to ECFP4 and FCFP4.
        length : int
            The length of the fingerprint. The default value is 2,048.
        features : bool
            If True, use pharmacophoric properties (FCFP) instead of atomic invariants (ECFP).
            The default value is False.
        type : {1, 2, 3}
            Define the type of the Morgan fingerprint function to be used, where:

            * ``1`` means **GetMorganFingerprintAsBitVect()**. It returns an explicit bit vector of size ``length`` (hashed fingerprint), where 0s and 1s represent the presence or absence of a given feature, respectively.
            * ``2`` means **GetHashedMorganFingerprint()**. It returns a sparse int vector ``length`` elements long (hashed fingerprint) containing the occurrence number of each feature.
            * ``3`` means **GetMorganFingerprint()**. It returns a sparse int vector 2^32 elements long containing the occurrence number of each feature.

            The default value is ``2``.

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        IllegalArgumentError
            If ``type`` is a value other than 1, 2, or 3.
        """
        try:
            if type == 1:
                func = AllChem.GetMorganFingerprintAsBitVect
                return func(self._rdmol, radius=radius,
                            nBits=length, useFeatures=features)
            elif type == 2:
                func = AllChem.GetHashedMorganFingerprint
                return func(self._rdmol, radius=radius,
                            nBits=length, useFeatures=features)
            elif type == 3:
                func = AllChem.GetMorganFingerprint
                return func(self._rdmol, radius=radius,
                            useFeatures=features)
            else:
                raise IllegalArgumentError("Informed type '%s' is invalid. "
                                           "Available options are 1, 2, or 3."
                                           % str(type))
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("Fingerprint could not be created.")

    def pharm2d_fp(self, sig_factory=None):
        """Generate a 2D pharmacophore fingerprint for the molecule ``mol``.

        Parameters
        ----------
        sig_factory : RDKit :class:`~rdkit.Chem.Pharm2D.SigFactory`, optional
            Factory object for producing signatures. The default signature factory is defined as shown below:

            >>> feat_factory = ChemicalFeatures.BuildFeatureFactory(MIN_FDEF_FILE)
            >>> sig_factory = SigFactory(feat_factory, minPointCount=2,
            ...                          maxPointCount=3, trianglePruneBins=False)
            >>> sig_factory.SetBins([(0, 2), (2, 5), (5, 8)])
            >>> sig_factory.Init()

        Raises
        ------
        FingerprintNotCreated
            If the fingerprint could not be created.
        """
        try:
            if sig_factory is None:
                sig_factory = _prepare_pharm2d_fp()["sigFactory"]
            return Generate.Gen2DFingerprint(self._rdmol, sig_factory)
        except Exception as e:
            logger.exception(e)
            raise FingerprintNotCreated("The fingerprint could not be created.")


def available_fp_functions():
    """Return a list of all fingerprints available at `FingerprintGenerator`."""
    regex = re.compile(".*([a-zA-Z]+)_fp", flags=0)
    funcs = list(filter(regex.match, dir(FingerprintGenerator)))
    return funcs


def _prepare_morgan_fp(fp_opt):
    params = {}
    if "radius" in fp_opt:
        params["radius"] = fp_opt["radius"]
    if "length" in fp_opt:
        params["length"] = fp_opt["length"]
    if "features" in fp_opt:
        params["features"] = fp_opt["features"]
    if "type" in fp_opt:
        params["type"] = fp_opt["type"]

    return params


def _prepare_pharm2d_fp(fp_opt=None):
    fp_opt = fp_opt or {}

    params = {}
    if "sigFactory" in fp_opt:
        sig_factory = fp_opt["sigFactory"]
    else:
        feat_factory = ChemicalFeatures.BuildFeatureFactory(MIN_FDEF_FILE)
        sig_factory = SigFactory(feat_factory, minPointCount=2,
                                 maxPointCount=3, trianglePruneBins=False)
        sig_factory.SetBins([(0, 2), (2, 5), (5, 8)])
        sig_factory.Init()

    params["sigFactory"] = sig_factory
    return params


def _prepare_fp_params(fp_function, fp_opt):
    if fp_function == "pharm2d_fp" and type(fp_opt) is dict:
        return _prepare_pharm2d_fp(fp_opt)
    elif fp_function == "morgan_fp" and type(fp_opt) is dict:
        return _prepare_morgan_fp(fp_opt)
    else:
        return {}


def generate_fp_for_mols(mols, fp_function=None, fp_opt=None, critical=False):
    """Generate molecular fingerprints for a sequence of molecules.

    Parameters
    ----------
    mols : iterable of :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
        A sequence of molecules.
    fp_function : str
        The fingerprint function to use. The default value is 'pharm2d_fp'.

        To check out the list of available functions, call the function :py:meth:`available_fp_functions`.
    fp_opt : dict, optional
        A set of parameters to pass to ``fp_function``.
    critical : bool
        If True, raises any exceptions caught during the generation of fingerprints.
        Otherwise, ignores all exceptions (the default). The error messages are always printed to the logging output.

    Returns
    -------
     : list of dict
        A list of dictionaries where each item contains the molecule name and its fingerprint.

        The dict is defined as follows:

            * ``mol`` (str): the molecule name;
            * ``fp`` (RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` \
            or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`): the fingerprint;

    Raises
    ------
    IllegalArgumentError
        If ``fp_function`` is not a function available in `FingerprintGenerator`.

    Examples
    --------

    First, let's define a set of molecules.

    >>> from luna.wrappers.base import MolWrapper
    >>> mols = [MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O"),
    ...         MolWrapper.from_smiles("C[C@@H](C(=O)O)N"),
    ...         MolWrapper.from_smiles("C1=CC(=CC=C1CC(C(=O)O)N)O")]

    Now, you can generate fingerprints for these molecules using the function :meth:`generate_fp_for_mols`.
    For example, let's create count ECFP4 fingerprints of size 1,024 for the above molecules.

    >>> from luna.mol.fingerprint import generate_fp_for_mols
    >>> fps = generate_fp_for_mols(mols, fp_function="morgan_fp", fp_opt={"length": 1024})

    Then, you can loop through the results as shown below:

    >>> for d in fps:
    >>>     print(f"{d['mol'].ljust(25)} - {len(d['fp'].GetNonzeroElements())}")
    N[C@@H](CCC(N)=O)C(O)=O   - 19
    C[C@@H](C(=O)O)N          - 12
    C1=CC(=CC=C1CC(C(=O)O)N)O - 24
    """

    funcs = available_fp_functions()
    if fp_function not in funcs:
        raise IllegalArgumentError("The fingerprint function is "
                                   "not available.")

    if fp_function is None:
        fp_function = "pharm2d_fp"
        logger.debug("No fingerprint function was defined. So, the default "
                     "fingerprint type will be used: 2D Pharmacophore "
                     "fingerprint.")

    logger.debug("Generating molecular fingerprints for %d molecules."
                 % len(mols))

    params = _prepare_fp_params(fp_function, fp_opt)
    fpg = FingerprintGenerator()

    fp_mols = []
    for idx, mol in enumerate(mols):
        try:
            fpg.mol = mol
            fp = getattr(fpg, fp_function)(**params)
            fp_mols.append({"fp": fp, "mol": mol.GetProp("_Name")})
        except Exception as e:
            logger.error("Molecule at position %d failed. Name: %s"
                         % (idx, mol.GetProp("_Name")))
            logger.exception(e)

            if critical:
                raise

    logger.debug("%d molecular sfingerprint(s) created." % len(fp_mols))

    return fp_mols



import configparser
import ast
import logging

logger = logging.getLogger()


class Config:
    """
    Parser for configuration files.

    Parameters
    ----------
    config_file : str
        The pathname for the configuration file.

    Attributes
    ----------

    config : `configparser.ConfigParser`
        The parsed configuration file.
    """

    def __init__(self, config_file):
        self.config = configparser.ConfigParser(allow_no_value=True)
        try:
            self.config.read(config_file)
        except Exception as e:
            logger.exception(e)
            raise IOError('Configuration file %s not read.' % config_file)

    def parse_value(self, section, property_name):
        value = self.config.get(section, property_name)
        return ast.literal_eval(value)

    def get_section_map(self, section):
        """
        Try to access the section ``section`` from the parsed configuration file.
        """
        section_map = {}
        options = self.config.options(section)
        for option in options:
            try:
                section_map[option] = self.config.get(section, option)
            except Exception as e:
                logger.exception(e)
                section_map[option] = None
        return section_map

    def as_dict(self, only_props=True):
        if only_props:
            return {k: v
                    for s in self.config.sections()
                    for k, v in self.config.items(s)}
        return {s: dict(self.config.items(s)) for s in self.config.sections()}

    def __getattr__(self, attr):
        if hasattr(self.config, attr):
            return getattr(self.config, attr)
        else:
            raise AttributeError("The attribute '%s' does not exist in the "
                                 "class %s." % (attr,
                                                self.__class__.__name__))



import warnings
import logging

from Bio.File import as_handle

from luna.MyBio.PDB.PDBParser import PDBParser
from luna.MyBio.PDB.PDBExceptions import PDBConstructionException
from luna.MyBio.PDB.PDBExceptions import PDBConstructionWarning
from luna.MyBio.PDB.StructureBuilder import StructureBuilder


logger = logging.getLogger()

PROBES_SET = ("ACD", "ACN", "ACT", "ADY", "AMN", "BDY", "BEN", "BUT", "CHX", "DFO", "DME", "EOL", "ETH", "PHN", "THS", "URE")


class FTMapParser(PDBParser):
    """Parse a PDB file generated by FTMap and return a Structure object."""

    def __init__(self, PERMISSIVE=True, QUIET=False,
                 FIX_ATOM_NAME_CONFLICT=False,
                 FIX_OBABEL_FLAGS=False,
                 FIX_ALL_OBABEL_ERRORS=False):
        """Create a FTMapParser object.

        The FTMapParser parser call a number of standard methods in an aggregated
        StructureBuilder object.

        Arguments:
         - PERMISSIVE - Evaluated as a Boolean. If false, exceptions in
           constructing the SMCRA data structure are fatal. If true (DEFAULT),
           the exceptions are caught, but some residues or atoms will be missing.
           THESE EXCEPTIONS ARE DUE TO PROBLEMS IN THE PDB FILE!.
         - structure_builder - an optional user implemented StructureBuilder class.
         - QUIET - Evaluated as a Boolean. If true, warnings issued in constructing
           the SMCRA data will be suppressed. If false (DEFAULT), they will be shown.
           These warnings might be indicative of problems in the PDB file!
        """
        self.PERMISSIVE = bool(PERMISSIVE)
        self.QUIET = bool(QUIET)

        if FIX_ALL_OBABEL_ERRORS:
            FIX_ATOM_NAME_CONFLICT = True
            FIX_OBABEL_FLAGS = True
        # MODBY: Alexandre Fassio
        # Correct conflicts in atom names.
        self.FIX_ATOM_NAME_CONFLICT = FIX_ATOM_NAME_CONFLICT
        # MODBY: Alexandre Fassio
        # Do not fix empty chains using the default method. Instead, FTMapParser implements its own method to fix empty chains.
        self.FIX_EMPTY_CHAINS = False
        # MODBY: Alexandre Fassio
        # Correct flags that were incorrectly set by OpenBabel.
        self.FIX_OBABEL_FLAGS = FIX_OBABEL_FLAGS
        # MODBY: Alexandre Fassio
        # Correct all OpenBabel errors.
        self.FIX_ALL_OBABEL_ERRORS = FIX_ALL_OBABEL_ERRORS

    def get_structure(self, id, file, only_clusters=None, only_compounds=None):
        """Return the structure.

        Arguments:
         - id - string, the id that will be used for the structure
         - file - name of the PDB file OR an open filehandle

        """
        with warnings.catch_warnings():
            if self.QUIET:
                warnings.filterwarnings("ignore", category=PDBConstructionWarning)

            # Make a StructureBuilder instance (pass id of structure as parameter)
            # Now it pass the PDB file as a parameter

            with as_handle(file, mode='rU') as handle:
                self._parse_chunks(handle.readlines(), id, file,
                                   only_clusters=only_clusters,
                                   only_compounds=only_compounds)

            # Return the Structure instance
            structure = self.structure_builder.get_structure()

        return structure

    def _parse_chunks(self, all_lines, id, file, only_clusters=None, only_compounds=None):

        chunks = []
        new_chunk = []

        self.only_clusters = set()
        if only_clusters:
            self.only_clusters = set(only_clusters)

        self.only_compounds = set()
        if only_compounds:
            self.only_compounds = self._get_valid_residue_ids(only_compounds)

        i = 0
        for i in range(0, len(all_lines)):
            line = all_lines[i]

            if line.startswith("HEADER"):
                if not line.startswith("HEADER    ") and line.startswith("HEADER "):
                    line = line.replace("HEADER ", "HEADER    ")

                if len(new_chunk) != 0:
                    chunks.append(new_chunk)
                    new_chunk = []
            new_chunk.append(line)

        if len(new_chunk) != 0:
            chunks.append(new_chunk)

        final_structure_builder = StructureBuilder()
        final_structure_builder.init_structure(id, file)
        final_structure_builder.init_model(0)

        curr_chain_id = "a"

        curr_serial_number = 1
        global_line_counter = 0
        for cid, chunk_lines in enumerate(chunks):
            self.structure_builder = StructureBuilder()
            self.structure_builder.init_structure(f"Chunk{cid}")

            self.header = None
            self.conects = None
            self.trailer = None
            self.line_counter = 0

            self._parse(chunk_lines)

            self.structure_builder.set_header(self.header)
            self.structure_builder.set_conects(self.conects)

            structure = self.structure_builder.get_structure()

            if structure.header["head"] == "protein":
                final_structure_builder.set_header(structure.header)

                if self.structure_builder.atom:
                    curr_serial_number = self.structure_builder.atom.serial_number + 1

            elif structure.header["head"].startswith("crosscluster.") and self.only_clusters:
                cluster_id = int(structure.header["head"].split(".")[1])
                if cluster_id not in self.only_clusters:
                    continue

            remapped_conects = {}
            for model_id in structure.child_dict:
                if not final_structure_builder.structure.has_id(model_id):
                    final_structure_builder.init_model(model_id)

                # Set the current model to be the one with id `model_id`, which can be an existing
                # model or a model just initialized.
                final_structure_builder.model = final_structure_builder.structure[model_id]

                empty_chains = []
                for chain_id in list(structure[model_id].child_dict):

                    # If some chain id is empty, then replace it with the current default chain id.
                    if chain_id.strip() == "":
                        structure[model_id][chain_id].id = curr_chain_id
                        chain_id = curr_chain_id
                        curr_chain_id = chr(ord(curr_chain_id) + 1)

                    if not final_structure_builder.model.has_id(chain_id):
                        final_structure_builder.model.add(structure[model_id][chain_id])

                        # Set the current chain to be the one given by the id `chain_id`.
                        final_structure_builder.chain = final_structure_builder.model[chain_id]

                        # Only modify residues if the structure contains a cluster of probes.
                        if structure.header["head"].startswith("crosscluster."):
                            remove_residues = []
                            # Loop over each residue and set probes to be considered ligands.
                            for res in final_structure_builder.chain:
                                # Modify probes to be hetatm.
                                if res.resname in PROBES_SET:
                                    res.id = ("H_%s" % res.resname, res.id[1], res.id[2])
                                    res.cluster_id = int(structure.header["head"].split(".")[1])

                                # Update the line number where the residue starts.
                                res.at_line = global_line_counter + res.at_line

                                # Ignore compounds that are not in the set ``only_compounds``.
                                if not self._accept_residue(res):
                                    remove_residues.append(res.id)
                                else:
                                    for atm in res.get_atoms():
                                        remapped_conects[atm.serial_number] = curr_serial_number
                                        atm.serial_number = curr_serial_number
                                        curr_serial_number += 1

                            # Remove residues not defined in ``only_compounds``.
                            for res_id in remove_residues:
                                del final_structure_builder.chain[res_id]
                    else:
                        # Set the current chain to be the one given by the id `chain_id`.
                        final_structure_builder.chain = final_structure_builder.model[chain_id]

                        # Loop over each residue and set probes to be considered ligands.
                        for res in structure[model_id][chain_id]:
                            # Only modify residues if the structure contains a cluster of probes.
                            if structure.header["head"].startswith("crosscluster."):
                                # Modify probes to be hetatm.
                                if res.resname in PROBES_SET:
                                    res.id = ("H_%s" % res.resname, res.id[1], res.id[2])
                                    res.cluster_id = int(structure.header["head"].split(".")[1])

                            # Update the line number where the residue starts.
                            res.at_line = global_line_counter + res.at_line

                            try:
                                # Ignore compounds that are not in the set ``only_compounds``.
                                if not self._accept_residue(res):
                                    continue
                                else:
                                    if curr_serial_number is not None:
                                        for atm in res.get_atoms():
                                            remapped_conects[atm.serial_number] = curr_serial_number
                                            atm.serial_number = curr_serial_number
                                            curr_serial_number += 1

                                final_structure_builder.chain.add(res)
                            except PDBConstructionException as message:
                                self._handle_PDB_exception(message, res.at_line)

                    if len(final_structure_builder.chain) == 0:
                        empty_chains.append(final_structure_builder.chain.id)

                # Remove empty chains.
                for chain_id in empty_chains:
                    del final_structure_builder.model[chain_id]

            # Remap the conects based on the atoms' new numbering.
            if remapped_conects:
                new_conects = {}
                for atm1, atoms in self.structure_builder.conects.items():
                    # Ignore atoms that were removed due to ``only_clusters`` and ``only_compounds``.
                    if atm1 not in remapped_conects:
                        continue

                    new_atoms_list = []
                    for atm2 in atoms:
                        # Ignore atoms that were removed due to ``only_clusters`` and ``only_compounds``.
                        if atm2 not in remapped_conects:
                            continue
                        new_atoms_list.append(remapped_conects[atm2])
                    new_conects[remapped_conects[atm1]] = new_atoms_list

                self.structure_builder.set_conects(new_conects)

                # Update the conects in final_structure_builder with the
                # remapped conects from the current structure
                final_structure_builder.conects.update(self.structure_builder.conects)
            elif structure.header["head"] == "protein":
                # Update the conects in final_structure_builder with the
                # conects from the original protein structure
                final_structure_builder.conects.update(self.structure_builder.conects)

            global_line_counter += self.structure_builder.line_counter

        self.structure_builder = final_structure_builder

    def _get_valid_residue_ids(self, res_ids):
        valid_residue_ids = set()

        for res_id in res_ids:
            if not isinstance(res_id, tuple) or len(res_id) not in [3, 4]:
                logger.warning("The residue id '%s' is not valid. "
                               "A valid residue id consists of a tuple containing three (hetflag, resseq, icode) or "
                               "four (structure id, model id, chain id, (hetflag, resseq, icode)) elements." % str(res_id))
                continue
            if len(res_id) == 4:
                if not isinstance(res_id[3], tuple) or len(res_id[3]) != 3:
                    logger.warning("The residue id '%s' is not valid. "
                                   "A valid residue id consists of a tuple containing three (hetflag, resseq, icode) or "
                                   "four (structure id, model id, chain id, (hetflag, resseq, icode)) elements." % str(res_id))
                    continue
            valid_residue_ids.add(res_id)
        return valid_residue_ids

    def _accept_residue(self, res):
        if self.only_compounds:
            for res_id in self.only_compounds:
                if len(res_id) == 4:
                    if res_id[1:] == res.get_full_id()[1:]:
                        return True
                elif len(res_id) == 3:
                    if res_id == res.id:
                        return True
            return False
        return True



import numpy as np
from openbabel import openbabel as ob

import logging
logger = logging.getLogger()


class AtomData:
    """Store atomic data (atomic number, coordinates, bond type, and serial number).

    Parameters
    ----------
    atomic_num : int
        Atomic number.
    coord : array_like of float (size 3)
        Atomic coordinates (x, y, z).
    bond_type : int
        Bond type.
    full_id : tuple
        The atom' full id as in :class:`~luna.MyBio.PDB.Atom.Atom`.
    serial_number : int, optional
        Atom serial number.

    Attributes
    ----------
    atomic_num : int
        The atomic number.
    bond_type : int
        The bond type.
    full_id : tuple
        The atom' full id
    serial_number : int or None
        The atom serial number.
    """

    def __init__(self, atomic_num, coord, bond_type, full_id=None, serial_number=None):
        self.atomic_num = atomic_num
        # Standardize all coordinate data to the same Numpy data type for consistence.
        self._coord = np.array(coord, "f")
        self.bond_type = bond_type
        self.full_id = full_id
        self.serial_number = serial_number

    @property
    def x(self):
        """float: The orthogonal coordinates for ``x`` in Angstroms."""
        return self._coord[0]

    @property
    def y(self):
        """float: The orthogonal coordinates for ``y`` in Angstroms."""
        return self._coord[1]

    @property
    def z(self):
        """float: The orthogonal coordinates for ``z`` in Angstroms."""
        return self._coord[2]

    @property
    def coord(self):
        """array_like of float (size 3): The atomic coordinates (x, y, z)."""
        return self._coord

    @coord.setter
    def coord(self, xyz):
        # Standardize all coordinate data to the same Numpy data type for consistence.
        self._coord = np.array(xyz, "f")

    def __repr__(self):
        full_atom_name = ""
        if self.full_id is not None:
            full_atom_name = "%s/%s/%s" % self.full_id[0:3]
            res_name = "%d%s" % (self.full_id[3][1],
                                 self.full_id[3][2].strip())
            atom_name = "%s" % self.full_id[4][0]

            if self.full_id[4][1] != " ":
                atom_name += "-%s" % self.full_id[4][1]
            full_atom_name += "/%s/%s" % (res_name, atom_name)

        return ("<ExtendedAtomData: atomic number=%d, coord=(%.3f, %.3f, %.3f), atom='%s', serial number=%s>"
                % (self.atomic_num, self.x, self.y, self.z, full_atom_name, str(self.serial_number)))

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(self, other.__class__):
            return (self.atomic_num == other.atomic_num
                    and np.all(self._coord == other._coord)
                    and self.full_id == other.full_id
                    and self.serial_number == other.serial_number)
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __hash__(self):
        """Overrides the default implementation"""
        return hash((self.atomic_num, tuple(self._coord),
                     self.full_id, self.bond_type,
                     self.serial_number))


class ExtendedAtom:
    """Extend :class:`~luna.MyBio.PDB.Atom.Atom` with additional properties and methods.

    Parameters
    ----------
    atom : :class:`~luna.MyBio.PDB.Atom.Atom`
        An atom.
    nb_info : iterable of `AtomData`, optional
        A sequence of `AtomData` containing information about atoms covalently bound to ``atom``.
    atm_grps : iterable of :class:`~luna.groups.AtomGroup`, optional
        A sequence of atom groups that contain ``atom``.
    invariants : list or tuple, optional
        Atomic invariants.
    """

    def __init__(self, atom, nb_info=None, atm_grps=None, invariants=None):
        self._atom = atom
        self._nb_info = nb_info or []
        self._atm_grps = atm_grps or []
        self._invariants = invariants

    @property
    def atom(self):
        """:class:`~luna.MyBio.PDB.Atom.Atom`, read-only."""
        return self._atom

    @property
    def neighbors_info(self):
        """list of `AtomData`, read-only: The list of `AtomData`
        containing information about atoms covalently bound to ``atom``.

        To add or remove neighbors information from ``neighbors_info`` use :py:meth:`add_nb_info`
        or :py:meth:`remove_nb_info`, respectively."""
        return self._nb_info

    @property
    def atm_grps(self):
        """list of :class:`~luna.groups.AtomGroup`, read-only: The list of atom groups that contain ``atom``.

        To add or remove atom groups from ``atm_grps`` use :py:meth:`add_atm_grps`
        or :py:meth:`remove_atm_grps`, respectively."""
        return self._atm_grps

    @property
    def invariants(self):
        """list: The list of atomic invariants."""
        return self._invariants

    @invariants.setter
    def invariants(self, invariants):
        self._invariants = invariants

    @property
    def atomic_num(self):
        """int, read-only: This atom's atomic number. This information is obtained from Open Babel."""
        return ob.GetAtomicNum(self.element)

    @property
    def electronegativity(self):
        """float, read-only: The Pauling electronegativity for this atom. This information is obtained from Open Babel."""
        return ob.GetElectroNeg(ob.GetAtomicNum(self.element))

    @property
    def full_id(self):
        """tuple, read-only: The full id of an atom is the tuple (structure id, model id, chain id,
        residue id, atom name, alternate location)."""
        return self._atom.get_full_id()

    @property
    def full_atom_name(self):
        """str, read-only: The full name of an atom is composed by the structure id, model id,
                chain id, residue name, residue id, atom name, and alternate location if available.
                Fields are slash-separated."""
        full_atom_name = "%s/%s/%s" % self.get_full_id()[0:3]
        res_name = "%s/%d%s" % (self._atom.parent.resname, self._atom.parent.id[1], self._atom.parent.id[2].strip())
        atom_name = "%s" % self._atom.name
        if self.altloc != " ":
            atom_name += "-%s" % self.altloc
        full_atom_name += "/%s/%s" % (res_name, atom_name)

        return full_atom_name

    def add_nb_info(self, nb_info):
        """ Add `AtomData` objects to ``neighbors_info``."""
        self._nb_info = list(set(self._nb_info + list(nb_info)))

    def add_atm_grps(self, atm_grps):
        """ Add :class:`~luna.groups.AtomGroup` objects to ``atm_grps``."""
        self._atm_grps = list(set(self._atm_grps + list(atm_grps)))

    def remove_nb_info(self, nb_info):
        """ Remove `AtomData` objects from ``neighbors_info``."""
        self._nb_info = list(set(self._nb_info) - set(nb_info))

    def remove_atm_grps(self, atm_grps):
        """ Remove :class:`~luna.groups.AtomGroup` objects from ``atm_grps``."""
        self._atm_grps = list(set(self._atm_grps) - set(atm_grps))

    def get_neighbor_info(self, atom):
        """Get information from a covalently bound atom."""
        for info in self._nb_info:
            if atom.get_full_id() == info.full_id:
                return info

    def is_neighbor(self, atom):
        """Check if a given atom is covalently bound to it."""
        return atom.get_full_id() in [i.full_id for i in self._nb_info]

    def as_json(self):
        """Represent the atom as a dict containing the structure id, model id,
           chain id, residue name, residue id, and atom name.

           The dict is defined as follows:

            * ``pdb_id`` (str): structure id;
            * ``model`` (str): model id;
            * ``chain`` (str): chain id;
            * ``res_name`` (str): residue name;
            * ``res_id`` (tuple): residue id (hetflag, sequence identifier, insertion code);
            * ``name`` (tuple): atom name (atom name, alternate location).

        """
        full_id = self.get_full_id()

        return {"pdb_id": full_id[0],
                "model": full_id[1],
                "chain": full_id[2],
                "res_name": self.parent.resname,
                "res_id": full_id[3],
                "name": full_id[4]}

    def __getattr__(self, attr):
        if hasattr(self._atom, attr):
            return getattr(self._atom, attr)
        else:
            raise AttributeError("The attribute '%s' does not exist in the class %s." % (attr, self.__class__.__name__))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        return "<ExtendedAtom: %s>" % self.full_atom_name

    def __sub__(self, other):
        # It calls __sub__() from Biopython.
        return self._atom - other._atom

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(self, other.__class__):
            return self.full_atom_name == other.full_atom_name
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __lt__(self, a2):
        # It substitutes the residue id for its index in order to keep the same order as in the PDB.
        full_id1 = self.full_id[0:2] + (self.parent.idx, ) + self.full_id[4:]
        full_id2 = a2.full_id[0:2] + (a2.parent.idx, ) + a2.full_id[4:]

        return full_id1 < full_id2

    def __hash__(self):
        """Overrides the default implementation"""
        return hash(self.full_atom_name)



from os import path
from collections import defaultdict

from luna.util.config import Config

import logging
logger = logging.getLogger()


class InteractionConfig(dict):
    """Generic class to define parameters for interactions.

    Parameters
    ----------
    config : dict, optional
        A dict containing parameters for calculating interactions.

    Examples
    --------

    >>> config = {"max_ha_dist_hb_inter": 2.5}
    >>> inter_config = InteractionConfig(config)
    >>> print(inter_config["max_ha_dist_hb_inter"])
    2.5
    inter_config["max_ha_dist_hb_inter"] = 3
    >>> print(inter_config["max_ha_dist_hb_inter"])
    3
    """

    def __init__(self, config=None, names_map=None):
        config = config or {}
        self.names_map = names_map or {}
        super().__init__(config)

    @property
    def params(self):
        """list: The list of parameters."""
        return sorted(self.keys())

    @classmethod
    def from_config_file(cls, config_file):
        if not path.exists(config_file):
            raise OSError("File '%s' does not exist." % config_file)

        config = {}
        params = Config(config_file)
        names_map = {}

        for section in params.sections():
            params_dict = params.get_section_map(section)
            params_dict = {prop: params.parse_value(section, prop)
                           for prop in params_dict}
            names_map.update({k: section for k in params_dict})
            config.update(params_dict)

        return cls(config, names_map)

    def save_config_file(self, config_file):
        args_by_section = defaultdict(list)
        for k, v in self.items():
            section = self.names_map.get(k, "Other")
            args_by_section[section].append((k, v))

        with open(config_file, "w") as OUT:
            for section in sorted(args_by_section):
                OUT.write("[%s]\n" % section)
                for arg, val in sorted(args_by_section[section]):
                    OUT.write("%s = %s\n" % (arg, val))
                OUT.write("\n")


class DefaultInteractionConfig(InteractionConfig):
    """Default parameters for calculating interactions in LUNA."""

    def __init__(self, **kwargs):

        file_path = path.abspath(path.join(path.realpath(__file__), '../'))
        default_config_file = f"{file_path}/config.cfg"

        config = InteractionConfig.from_config_file(default_config_file)

        for k, v in kwargs.items():
            config[k] = v

        super().__init__(config)



import logging

logger = logging.getLogger()


class VersionControl():

    def __init__(self):
        from MyBio.PDB.PDBList import PDBList
        from urllib.error import URLError, HTTPError

        pdbl = PDBList()
        try:
            new, modified, obsolete = pdbl.get_recent_changes()
        except HTTPError as e:
            logger.exception(e)
            raise HTTPError("PDB server could not be reached.")
        except URLError as e:
            logger.exception(e)
            raise URLError("PDB server could not be reached.")

        self.new = set(new)
        self.modified = set(modified)
        self.obsolete = set(obsolete)

    def is_pdb_outdated(self, pdbId):
        if (pdbId in self.new or pdbId in self.modified):
            return True
        else:
            return False

    def is_pdb_obsolete(self, pdbId):
        if (pdbId in self.obsolete):
            return True
        else:
            return False



from itertools import product

from luna.util.default_values import COV_SEARCH_RADIUS, BOUNDARY_CONFIG
from luna.util.exceptions import EntityLevelError, IllegalArgumentError
from luna.MyBio.util import ENTITY_LEVEL_NAME, is_covalently_bound
from luna.MyBio.PDB.NeighborSearch import NeighborSearch
from luna.MyBio.PDB import Selection
from luna.MyBio.PDB.Residue import Residue

import logging

logger = logging.getLogger()


def get_all_contacts(entity, radius=BOUNDARY_CONFIG["bsite_cutoff"],
                     level='A'):
    """Recover all residue-residue or atom-atom contacts in ``entity``.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object from where atoms and residues will be recovered.
    radius : float
        The cutoff distance (in Å) for defining contacts. The default value is 6.2.
    level : {'R', 'A'}
        Return residues ('R') or atoms ('A') in contact with ``source``.

    Returns
    -------
     : list of tuple of (:class:`~luna.MyBio.PDB.Residue.Residue` or :class:`~luna.MyBio.PDB.Atom.Atom`, \
      :class:`~luna.MyBio.PDB.Residue.Residue` or :class:`~luna.MyBio.PDB.Atom.Atom`)
        Each tuple contains either a pair of residues or atoms in contact.

    Raises
    ------
    EntityLevelError
        If ``level`` is neither 'R' nor 'A'.

    Examples
    --------

    In this example, we will identify all residue-residue contacts within 2.5 Å in the PDB 3QQK.
    So, let's first parse the PDB file.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", f"{LUNA_PATH}/tutorial/inputs/3QQK.pdb")

    Then, to recover all residue-residue contacts within 2.5 Å use :meth:`get_all_contacts`
    with ``level`` set to 'R' and radius set to 2.5.

    >>> from luna.interaction.contact import get_all_contacts
    >>> contacts = get_all_contacts(structure, radius=2.5, level="R")
    >>> print(len(contacts))
    314

    """
    try:
        if level == 'S' or level == 'M' or level == 'C':
            raise EntityLevelError("Maximum entity level to be chosen is: R (residues) or A (atoms)")

        if level not in ENTITY_LEVEL_NAME:
            raise EntityLevelError("The defined level '%s' does not exist" % level)

        logger.debug("Trying to select all contacts in the PDB file %s." % entity.get_parent_by_level('S').id)

        all_atoms = list(entity.get_atoms())
        ns = NeighborSearch(all_atoms)
        pairs = ns.search_all(radius, level)

        logger.debug("Number of nearby %s(s) found: %d." % (ENTITY_LEVEL_NAME[level].lower(), len(pairs)))

        return pairs
    except Exception as e:
        logger.exception(e)
        raise


def get_contacts_with(entity, source, target=None,
                      radius=BOUNDARY_CONFIG["bsite_cutoff"], level='A'):
    """Recover atoms or residues in contact with ``source``.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object from where atoms and residues will be recovered.
    source : :class:`~luna.MyBio.PDB.Entity.Entity`
        The reference, which can be any :class:`~luna.MyBio.PDB.Entity.Entity` instance (structure, model, chain, residue, or atom).
    target : :class:`~luna.MyBio.PDB.Entity.Entity`, optional
        If provided, only contacts with the ``target`` will be considered.
    radius : float
        The cutoff distance (in Å) for defining contacts. The default value is 6.2.
    level : {'R', 'A'}
        Return residues ('R') or atoms ('A') in contact with ``source``.

    Returns
    -------
     : set of tuple of (:class:`~luna.MyBio.PDB.Residue.Residue` or :class:`~luna.MyBio.PDB.Atom.Atom`, \
    :class:`~luna.MyBio.PDB.Residue.Residue` or :class:`~luna.MyBio.PDB.Atom.Atom`)
        Each tuple contains two items:\
            the first corresponds to a residue/atom from the ``source``,\
            and the second corresponds to a residue/atom in contact with ``source``.

    Raises
    ------
    EntityLevelError
        If ``level`` is neither 'R' nor 'A'.

    Examples
    --------

    **Example 1)** In this example, we will identify residue-residue contacts between a ligand and nearby residues.

    First, let's parse a PDB file to work with.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", f"{LUNA_PATH}/tutorial/inputs/3QQK.pdb")

    Now, we define the target ligand.

    >>> ligand = structure[0]["A"][('H_X02', 497, ' ')]

    Then, to recover contacts between this ligand and nearby residues we use :meth:`get_contacts_with`
    with ``level`` set to 'R'.

    >>> from luna.interaction.contact import get_contacts_with
    >>> contacts = sorted(get_contacts_with(structure, ligand, radius=3, level="R"))
    >>> for pair in contacts:
    ...     print(pair)
    (<Residue X02 het=H_X02 resseq=497 icode= >, <Residue GLU het=  resseq=81 icode= >)
    (<Residue X02 het=H_X02 resseq=497 icode= >, <Residue LEU het=  resseq=83 icode= >)
    (<Residue X02 het=H_X02 resseq=497 icode= >, <Residue X02 het=H_X02 resseq=497 icode= >)
    (<Residue X02 het=H_X02 resseq=497 icode= >, <Residue HOH het=W resseq=321 icode= >)


    **Example 2)** In this example, we will identify atom-atom contacts between a ligand and a given residue.

    First, let's parse a PDB file to work with.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", f"{LUNA_PATH}/tutorial/inputs/3QQK.pdb")

    Now, we define the target ligand and residue.

    >>> ligand = structure[0]["A"][('H_X02', 497, ' ')]
    >>> residue = structure[0]["A"][(' ', 81, ' ')]

    Then, to recover contacts between these compounds we use :meth:`get_contacts_with`.
    As we need atom-wise contacts, set ``level`` to 'A'.

    >>> from luna.interaction.contact import get_contacts_with
    >>> contacts = sorted(get_contacts_with(structure, ligand, target=residue, radius=4, level="A"))
    >>> for pair in contacts:
    ...     print(pair)
    (<Atom C7>, <Atom O>)
    (<Atom N10>, <Atom C>)
    (<Atom N10>, <Atom O>)
    """
    try:
        if level == 'S' or level == 'M' or level == 'C':
            raise EntityLevelError("Maximum entity level to be chosen is: "
                                   "R (residues) or A (atoms)")

        if level not in ENTITY_LEVEL_NAME:
            raise EntityLevelError("The defined level '%s' does not exist"
                                   % level)

        source_atoms = Selection.unfold_entities([source], 'A')
        target_atoms = []
        if target is None:
            target_atoms = list(entity.get_atoms())
        else:
            target_atoms = Selection.unfold_entities(target, 'A')

        ns = NeighborSearch(target_atoms)
        entities = set()
        for atom in source_atoms:
            entity = atom.get_parent_by_level(level)
            nb_entities = ns.search(atom.coord, radius, level)
            pairs = set(product([entity], nb_entities))
            entities.update(pairs)

        logger.debug("Number of nearby %s(s) found: %d."
                     % (ENTITY_LEVEL_NAME[level].lower(), len(entities)))
        return entities
    except Exception as e:
        logger.exception(e)
        raise


def get_cov_contacts_with(entity, source, target=None):
    """Recover potential covalent bonds with ``source``.

    Covalent bonds between two nearby atoms A and B are determined as in Open Babel:

    .. math:: 0.4 <= \overrightarrow{\|AB\|} <= A_{cov} + B_{cov} + 0.45

    Where :math:`\overrightarrow{\|AB\|}` is the distance between atoms A and B,
    and :math:`A_{cov}` and :math:`B_{cov}` are the covalent radius of atoms A and B, respectively.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object from where atoms will be recovered.
    source : :class:`~luna.MyBio.PDB.Entity.Entity`
        The reference, which can be any :class:`~luna.MyBio.PDB.Entity.Entity` instance (structure, model, chain, residue, or atom).
    target : :class:`~luna.MyBio.PDB.Entity.Entity`, optional
        If provided, only covalent bonds with the ``target`` will be considered.

    Returns
    -------
     : set of tuple of (:class:`~luna.MyBio.PDB.Atom.Atom`, :class:`~luna.MyBio.PDB.Atom.Atom`)
        Pairs of atoms with potential covalent bonds.


    Examples
    --------

    In this example, we will identify all covalent bonds involving a given residue in the PDB 3QQK.
    So, let's first parse the PDB file.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", f"{LUNA_PATH}/tutorial/inputs/3QQK.pdb")

    Now, we select the residue of our interest.

    >>> residue = structure[0]["A"][(' ', 81, ' ')]

    Finally, let`s recover potential covalent bonds involving this residue with :meth:`get_cov_contacts_with`.
    In the below snippet, pairs of atoms are sorted by atoms` serial number,
    and the residue information is printed together with the atom name.

    >>> from luna.interaction.contact import get_cov_contacts_with
    >>> cov_bonds = sorted(get_cov_contacts_with(structure, residue), key=lambda x: (x[0].serial_number, x[1].serial_number))
    >>> for atm1, atm2 in cov_bonds:
    >>>     pair = ("%s%d/%s" % (atm1.parent.resname, atm1.parent.id[1], atm1.name),
    ...             "%s%d/%s" % (atm2.parent.resname, atm2.parent.id[1], atm2.name))
    >>>     print(pair)
    ('PHE80/C', 'GLU81/N')
    ('GLU81/N', 'GLU81/CA')
    ('GLU81/CA', 'GLU81/C')
    ('GLU81/CA', 'GLU81/CB')
    ('GLU81/C', 'GLU81/O')
    ('GLU81/C', 'PHE82/N')
    ('GLU81/CB', 'GLU81/CG')
    ('GLU81/CG', 'GLU81/CD')
    ('GLU81/CD', 'GLU81/OE1')
    ('GLU81/CD', 'GLU81/OE2')
    """
    entities = get_contacts_with(entity, source, target, radius=COV_SEARCH_RADIUS, level='A')

    cov_bonds = set()
    for atm1, atm2 in entities:
        if is_covalently_bound(atm1, atm2):
            cov_bonds.add(tuple(sorted([atm1, atm2], key=lambda x: x.serial_number)))

    return cov_bonds


def get_proximal_compounds(source, radius=COV_SEARCH_RADIUS):
    """Recover proximal compounds to ``source``.

    Parameters
    ----------
    source : :class:`~luna.MyBio.PDB.Residue.Residue`
        The reference compound.
    radius : float
        The cutoff distance (in Å) for defining proximity.\
        The default value is 2.2, which may recover potential residues bound to ``source`` through covalent bonds.

    Returns
    -------
     : list of :class:`~luna.MyBio.PDB.Residue.Residue`
        The list of proximal compounds always include ``source``.

    Raises
    ------
    IllegalArgumentError
        If ``source`` is not a :class:`~luna.MyBio.PDB.Residue.Residue`

    Examples
    --------

    In this example, we will identify all proximal compounds to a given residue in the PDB 3QQK.
    So, let's first parse the PDB file.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", f"{LUNA_PATH}/tutorial/inputs/3QQK.pdb")

    Now, we select the residue of our interest.

    >>> residue = structure[0]["A"][(' ', 81, ' ')]

    Finally, we call :meth:`get_proximal_compounds`.

    >>> from luna.interaction.contact import get_proximal_compounds
    >>> compounds = get_proximal_compounds(residue)
    >>> for c in compounds:
    ...    print(c)
    <Residue PHE het=  resseq=80 icode= >
    <Residue GLU het=  resseq=81 icode= >
    <Residue PHE het=  resseq=82 icode= >
    """
    if not isinstance(source, Residue):
        raise IllegalArgumentError("Invalid object type. A Residue object is expected instead.")

    model = source.get_parent_by_level('M')
    proximal = get_contacts_with(entity=model, source=source, radius=radius, level='R')

    # Sorted by the compound order as in the PDB.
    return sorted(list(set([p[1] for p in proximal])), key=lambda r: (r.parent.parent.id, r.parent.id, r.idx))



import multiprocessing as mp
import time
from collections import Sequence

from luna.util.progress_tracker import ProgressData, ProgressTracker
from luna.util.file import new_unique_filename

import logging
logger = logging.getLogger()


MAX_NPROCS = mp.cpu_count() - 1


class Sentinel:
    """Custom sentinel to stop workers"""
    pass


class ArgsGenerator:
    """Custom generator that implements __len__().
       This class can be used in conjunction with :class:`~luna.util.progress_tracker.ProgressTracker` in cases
       where the tasks are obtained from generators. Note that :class:`~luna.util.progress_tracker.ProgressTracker`
       requires a pre-defined number of tasks to calculate the progress, therefore a standard generator cannot be
       used directly as it does not implement __len__(). Then, with `ArgsGenerator`, one may take advantage of
       generators and :class:`~luna.util.progress_tracker.ProgressTracker` by explicitly providing the number
       of tasks that will be generated.

       Parameters
       ----------
       generator : generator
           The tasks generator.
       nargs : int
           The number of tasks that will be generated.
    """

    def __init__(self, generator, nargs):
        self.generator = generator
        self.nargs = nargs

    def __len__(self):
        return self.nargs

    def __iter__(self):
        for d in self.generator:
            yield d


class ParallelJobs:

    """Executes a set of tasks in parallel (:py:class:`~multiprocessing.JoinableQueue`) or sequentially.

    Parameters
    ----------
    nproc : int or None
       The number of CPUs to use. The default value is the ``maximum number of CPUs - 1``.
       If ``nproc`` is None, 0, or 1, run the jobs sequentially. Otherwise, use the ``maximum number of CPUs - 1``.

    Attributes
    ----------
    nproc : int
        The number of CPUs to use.
    progress_tracker : ProgressTracker
        A :class:`~luna.util.progress_tracker.ProgressTracker` object to track the tasks' progress.
    """

    # TODO: add option to Threads/Multiprocessing
    def __init__(self, nproc=MAX_NPROCS):

        if nproc is not None:
            # Use 'MAX_NPROCS' if a non-integer has been provided.
            if not isinstance(nproc, int) or nproc < 0:
                nproc = MAX_NPROCS
            elif nproc in [0, 1]:
                nproc = None
            else:
                nproc = min(nproc, MAX_NPROCS)

        self.nproc = nproc
        self.progress_tracker = None

    def _exec_func(self, data, func):
        start = time.time()

        output = None
        exception = None
        try:
            # Unpack args as a dictionary.
            if isinstance(data, dict):
                output = func(**data)
            # Unpack args as a list.
            elif isinstance(data, Sequence):
                output = func(*data)
            # Try unpack the function as a single arg.
            else:
                output = func(data)
        # Capture any errors while executing the above code.
        except Exception as e:
            logger.exception(e)
            exception = e

        proc_time = time.time() - start

        return output, exception, proc_time

    def _producer(self, args, job_queue):
        for data in args:
            job_queue.put(data)

    def _consumer(self, func, job_queue, progress_queue, output_queue=None):
        while True:
            data = job_queue.get()

            # If sentinel is found, break.
            if isinstance(data, Sentinel):
                break

            # Execute the provided function.
            output, exception, proc_time = self._exec_func(data, func)

            if output is not None and output_queue is not None:
                output_queue.put((data, output))

            # Update progress tracker.
            pd = ProgressData(input_data=data, output_data=output, exception=exception, proc_time=proc_time, func=func)
            progress_queue.put(pd)

            job_queue.task_done()

    def _saver(self, output_queue, output_file, proc_func=None, output_header=None):

        with open(output_file, "w") as OUT:
            if output_header is not None:
                OUT.write(output_header.strip())
                OUT.write("\n")

            while True:
                data = output_queue.get()

                # If sentinel is found, break.
                if isinstance(data, Sentinel):
                    break

                line = None
                if proc_func is not None:
                    # Execute the provided function.
                    output, exception, proc_time = self._exec_func(data,
                                                                   proc_func)
                    line = output

                try:
                    # If no data is stored in line, try to access the
                    # output generated by the _consumer() function.
                    if line is None:
                        line = data[1]

                    OUT.write(str(line).strip())
                    OUT.write("\n")
                    OUT.flush()
                except Exception as e:
                    logger.error("An error occurred while trying to save "
                                 "the output '%s'." % str(line))
                    logger.exception(e)

                output_queue.task_done()

    def _sequential(self, args, func, progress_queue):
        # Run jobs sequentially.
        for data in args:
            # Execute provided function.
            output, exception, proc_time = self._exec_func(data, func)

            # Save data.
            pd = ProgressData(input_data=data, output_data=output, exception=exception, proc_time=proc_time)

            # Update progress tracker.
            progress_queue.put(pd)

    def run_jobs(self, args, consumer_func, output_file=None, proc_output_func=None, output_header=None, job_name=None):
        """
        Run a set of tasks in parallel or sequentially according to the ``nproc``.

        Parameters
        ----------
        args : iterable of iterables, `ArgsGenerator`
            A sequence of arguments to be provided to the consumer function ``consumer_func``.
        consumer_func : function
            The function that will be executed for each set of arguments in ``args``.
        output_file : str, optional
            Save outputs to this file.
            If ``proc_output_func`` is not provided, it tries to save a stringified version of each output data.
            Otherwise, it executes ``proc_output_func`` first and its output will be printed to the output file instead.

            Note: if ``proc_output_func`` is provided but not ``output_file``, a new random unique filename will
            be generated and the file will be saved in the current directory.
        proc_output_func : function, optional
            Post-processing function that is executed for each output data produced by ``consumer_func``.
        output_header : str, optional
            A header for the output file.
        job_name : str, optional
            A name to identify the job.

        Returns
        -------
         : :class:`~luna.util.progress_tracker.ProgressResult`
        """
        if proc_output_func is not None and output_file is None:
            output_file = new_unique_filename(".") + ".output"
            logger.warning("No output file was defined. So, it will try to save results at '%s'." % output_file)
        elif output_file is not None:
            logger.warning("The output file '%s' was defined. So, it will try to save results at it." % output_file)

        # Queue for progress tracker.
        progress_queue = mp.JoinableQueue(maxsize=1)

        # Progress tracker
        self.progress_tracker = ProgressTracker(len(args), progress_queue,
                                                job_name)
        self.progress_tracker.start()

        # Initialize a new progress bar (display a 0% progress).
        progress_queue.put(None)

        if self.nproc is not None:
            job_queue = mp.JoinableQueue(maxsize=self.nproc)

            output_queue = None
            if output_file is not None:
                output_queue = mp.JoinableQueue()

            for i in range(self.nproc):
                p = mp.Process(name="ConsumerProcess-%d" % i,
                               target=self._consumer,
                               args=(consumer_func, job_queue,
                                     progress_queue, output_queue,))
                p.daemon = True
                p.start()

            if output_file is not None:
                o = mp.Process(name="WriterProcess-%d" % i,
                               target=self._saver,
                               args=(output_queue, output_file,
                                     proc_output_func, output_header,))
                o.daemon = True
                o.start()

            # Produce tasks to consumers.
            self._producer(args, job_queue)

            # Sentinels to stop consumers.
            sentinel = Sentinel()

            # Join all processes and add sentinels to stop consumers.
            job_queue.join()
            [job_queue.put(sentinel) for i in range(self.nproc)]

            if output_file is not None:
                output_queue.join()
                output_queue.put(sentinel)

        else:
            self._sequential(args, consumer_func, progress_queue)

        # Finish the progress tracker.
        self.progress_tracker.end()

        return self.progress_tracker.results



import re
import logging
from operator import xor
from os.path import exists
from collections import defaultdict
import ast

from rdkit.Chem import Mol as RDMol
from openbabel import OBMol
from openbabel.pybel import readfile
from openbabel.pybel import Molecule as PybelMol
from openbabel.pybel import informats as OB_FORMATS

from luna.wrappers.rdkit import RDKIT_FORMATS, read_multimol_file, read_mol_from_file
from luna.wrappers.base import MolWrapper
from luna.util.default_values import ACCEPTED_MOL_OBJ_TYPES, ENTRY_SEPARATOR, ARTIFACTS_LIST
from luna.util.file import get_file_format, get_filename
from luna.util.exceptions import InvalidEntry, IllegalArgumentError, MoleculeObjectError, MoleculeObjectTypeError, MoleculeNotFoundError
from luna.MyBio.PDB.PDBParser import PDBParser, WATER_NAMES, DEFAULT_CHAIN_ID
from luna.MyBio.PDB.Entity import Entity


logger = logging.getLogger()

PCI_ENTRY_REGEX = re.compile(r'^.{1,255}:\w:\w[\w+\-]{0,2}:\-?\d{1,4}[a-zA-z]?$')
PPI_ENTRY_REGEX = re.compile(r'^.{1,255}:\w$')

REGEX_RESNUM_ICODE = re.compile(r'^([\-\+]?\d+)([a-zA-z]?)$')


class Entry:
    """Entries determine the target molecule to which interactions and other properties
    will be calculated. They can be ligands, chains, etc, and can be defined in a number of ways.
    Each entry has an associated PDB file that may contain macromolecules (protein, RNA, DNA) and
    other small molecules, water, and ions.
    The PDB file provides the context to where the interactions with the target molecule will be calculated.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    comp_name : str, optional
        A 1 to 3-symbols compound name (residue name in the PDB format).
        Obligatory if ``is_hetatm`` is True. Example: 'X01'.
    comp_num : int, optional
        A valid 4-digits integer (residue sequence number in the PDB format).
        Obligatory if ``is_hetatm`` is True. Example: 300 or -1.
    comp_icode : str, optional
        A 1-character compound insertion code (residue insertion code in the PDB format). Example: 'A'.
    is_hetatm : bool
        If the compound is a ligand or not. The default value is True.
    sep : str
        A separator character to format the entry string. The default value is ':'.
    parser : :class:`~luna.MyBio.PDB.PDBParser.PDBParser` or :class:`~luna.MyBio.PDB.FTMapParser.FTMapParser`, optional
            Define a PDB parser object. If not provided, the default parser will be used.

    Raises
    ------
    IllegalArgumentError
        If ``is_hetatm`` is True, but the compound name and number are not provided.
        If the compound number is provided but it is not an integer.
        If ``comp_icode`` is provided but it is not a valid character.
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------

    Chain entry: can be used to calculate interactions with a given chain.

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A")
    >>> print(e)
    <Entry: 3QL8:A>

    Compound entry: can be used to calculate interactions with a given compound (residue or nucleotide).

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="HIS", comp_num=125, is_hetatm=False)
    >>> print(e)
    <Entry: 3QL8:A:HIS:125>

    Ligand entry: can be used to calculate interactions with a given ligand.

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, is_hetatm=True)
    >>> print(e)
    <Entry: 3QL8:A:X01:300>

    You can use a different character separator for the entries. For example:

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, is_hetatm=True, sep="/")
    >>> print(e)
    <Entry: 3QL8/A/X01/300>
    """

    def __init__(self, pdb_id, chain_id, comp_name=None,
                 comp_num=None, comp_icode=None,
                 is_hetatm=True, sep=ENTRY_SEPARATOR, parser=None):

        if xor(comp_name is None, comp_num is None):
            raise IllegalArgumentError("You tried to define a compound, so "
                                       "you must inform its name and number.")

        if comp_num is not None:
            try:
                assert float(comp_num).is_integer()
                comp_num = int(comp_num)
            except (ValueError, AssertionError):
                raise IllegalArgumentError("The informed compound number '%s' is invalid. It must be an integer." % str(comp_num))

        if comp_icode is not None:
            comp_icode = str(comp_icode)
            if comp_icode.isdigit() or len(comp_icode) > 1:
                raise IllegalArgumentError("The informed compound icode '%s' is invalid. It must be a character." % str(comp_icode))

        self._pdb_id = pdb_id
        self._chain_id = chain_id
        self._comp_name = comp_name
        self._comp_num = comp_num
        self._comp_icode = comp_icode
        self.is_hetatm = is_hetatm
        self.sep = sep
        self.parser = parser

        if not self.is_valid():
            raise InvalidEntry("Entry '%s' does not match the PDB format."
                               % self.to_string())

    @classmethod
    def from_string(cls, entry_str, is_hetatm=True, sep=ENTRY_SEPARATOR):
        """Initialize from a string.

        Parameters
        ----------
        entry_str : str
            A string representing the entry. Example: '3QL8:A:X01:300'.
        is_hetatm : bool
            Defines if the compound is a ligand or not. The default value is True.
        sep : str
            The separator character used in ``entry_str``. The default value is ':'.
            For example: if ``entry_str`` is set to '3QL8|A|X01|300', then ``sep`` should be defined as '|'.

        Returns
        -------
         : `Entry`

        Raises
        ------
        IllegalArgumentError
            If the fields in ``entry_str`` do not match the format expected to define a
            chain (ChainEntry) or a compound (MolEntry).

        Examples
        --------

        Chain entry: can be used to calculate interactions with a given chain.

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A", sep=":")
        >>> print(e)
        <Entry: 3QL8:A>

        Compound entry: can be used to calculate interactions with a given compound (residue or nucleotide).

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A:HIS:125", sep=":")
        >>> print(e)
        <Entry: 3QL8:A:HIS:125>

        Ligand entry: can be used to calculate interactions with a given ligand.

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A:X01:300", sep=":")
        >>> print(e)
        <Entry: 3QL8:A:X01:300>
        """

        entries = entry_str.split(sep)

        # Try to initialize a new ChainEntry.
        if len(entries) == 2:
            if any([str(i).strip() == "" for i in entries]):
                raise IllegalArgumentError("The number of fields in the informed string '%s' is incorrect. A valid ChainEntry must contain "
                                           "two obligatory fields: PDB and chain id." % entry_str)

            return cls(*entries, is_hetatm=False, sep=sep)

        # Try to initialize a new MolEntry.
        elif len(entries) == 4:
            if any([str(i).strip() == "" for i in entries]):
                raise IllegalArgumentError("The number of fields in the informed string '%s' is incorrect. A valid MolEntry "
                                           "must contain four obligatory fields: PDB, chain id, compound name, and compound "
                                           "number followed by its insertion code when applicable." % entry_str)

            # Separate ligand number from insertion code.
            matched = REGEX_RESNUM_ICODE.match(entries[3])
            if matched:
                comp_num = matched.group(1)
                try:
                    assert float(comp_num).is_integer()
                    comp_num = int(comp_num)
                except (ValueError, AssertionError):
                    raise IllegalArgumentError("The informed compound number '%s' is invalid. It must be an integer." % str(comp_num))

                icode = None if matched.group(2) == "" else matched.group(2)
                entries = entries[0:3] + [comp_num, icode]
            else:
                raise IllegalArgumentError("The compound number and its insertion code (if applicable) '%s' is invalid. "
                                           "It must be an integer followed by one insertion code character when applicable."
                                           % entries[3])

            return cls(*entries, is_hetatm=is_hetatm, sep=sep)
        else:
            raise IllegalArgumentError("The number of fields in the informed string '%s' is incorrect. A valid string must contain "
                                       "two obligatory fields (PDB and chain id) and it may contain two optional fields (compound name "
                                       "and compound number followed by its insertion code when applicable)." % entry_str)

    @classmethod
    def from_file(cls, input_file, pdb_id=None, mol_file=None,
                  entries_sep=":", fields_sep=",", **kwargs):
        #     """Initialize from a list of strings representing compounds.

        #     Parameters
        #     ----------
        #     input_file : str
        #         The file from where the list of strings (one per line) will be read from.
        #     sep : str
        #         The separator character used in ``input_file``. The default value is ':'.
        #         For example: if entries from ``input_file`` use '|' as the separator, then ``sep`` should be defined as '|'.

        if not exists(input_file):
            raise OSError("File '%s' does not exist." % input_file)

        with open(input_file, "r") as IN:
            c = 1
            for row in IN:
                row = row.strip()
                if row == "":
                    continue

                args = row.split(fields_sep)

                has_error = False
                if len(args) == 1:
                    args = row.split(entries_sep)

                    if len(args) == 1:
                        if pdb_id is None or mol_file is None:
                            msg = ("It seems a molecule name was provided "
                                   f"in line #{c}. In this case, "
                                   "'pdb_id' and 'mol_file' are mandatory.")
                            raise IllegalArgumentError(msg)

                        yield MolFileEntry.from_mol_file(pdb_id, args[0],
                                                         mol_file,
                                                         is_multimol_file=True,
                                                         **kwargs)

                    elif len(args) == 2:
                        yield ChainEntry.from_string(row, sep=entries_sep)

                    elif len(args) == 4:
                        yield MolEntry.from_string(row, sep=entries_sep)

                    else:
                        has_error = True

                elif len(args) == 4:
                    (curr_pdb_id, curr_mol_id,
                        curr_mol_file, is_multimol) = args
                    is_multimol = ast.literal_eval(is_multimol)

                    cls = MolFileEntry
                    yield cls.from_mol_file(curr_pdb_id, curr_mol_id,
                                            curr_mol_file,
                                            is_multimol_file=is_multimol,
                                            **kwargs)
                else:
                    has_error = True

                if has_error:
                    msg = (f"Invalid number of arguments found in line #{c}. "
                           "You should provide either one or four arguments. "
                           "If only one argument is provided, it can be an "
                           "entry string (e.g., 3QQK:A:X02:497) or a "
                           "molecule name.")
                    raise IllegalArgumentError(msg)

                c += 1

    @property
    def pdb_id(self):
        """str, read-only: the pdb id."""
        return self._pdb_id

    @property
    def chain_id(self):
        """str, read-only: the chain id."""
        return self._chain_id

    @property
    def comp_name(self):
        """str, read-only: the compound name."""
        return self._comp_name

    @property
    def comp_num(self):
        """int, read-only: the compound number."""
        return self._comp_num

    @property
    def comp_icode(self):
        """str, read-only: the compound insertion code."""
        if isinstance(self._comp_icode, str):
            return self._comp_icode
        else:
            return ' '

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple (PDB id or filename, chain id)
        for entries representing chains and (PDB id or filename, chain id,
        compound name, compound number, insertion code) for entries representing compounds."""

        entry = [self.pdb_id, self.chain_id]
        if self.comp_name is not None and self.comp_num is not None:
            entry.append(self.comp_name)
            entry.append(self.comp_num)
            entry.append(self.comp_icode)
        return tuple(entry)

    def to_string(self, sep=None):
        """ Convert the entry to a string using ``sep`` as a separator character.

        Parameters
        ----------
        sep : str or None
            If None (the default), use the separator character defined during the entry object creation.
            Otherwise, uses ``sep`` as the separator character.

        Examples
        --------

        >>> from luna.mol.entry import Entry
        >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, is_hetatm=True, sep=":")
        >>> print(e.to_string("/"))
        3QL8/A/X01/300
        """

        full_id = self.full_id

        # An entry object will always have a PDB and chain id.
        entry = list(full_id[0:2])

        # If it contains additional information about the compound it will also include them.
        if len(full_id) > 2:
            if full_id[2] is not None and full_id[3] is not None:
                comp_name = str(full_id[2]).strip()
                comp_num_and_icode = str(full_id[3]).strip() + str(full_id[4]).strip()
                entry += [comp_name, comp_num_and_icode]

        sep = sep or self.sep

        return sep.join(entry)

    def is_valid(self):
        """Check if the entry matches the expected format for protein-protein or protein-compound complexes.

        Returns
        -------
         : bool
        """
        full_id = self.full_id

        # Regex for ChainEntry (pdb_id, chain_id).
        if len(full_id) == 2:
            return PPI_ENTRY_REGEX.match(self.to_string(":")) is not None

        # Regex for MolEntry (pdb_id, chain_id, comp_name, comp_num, icode).
        elif len(full_id) == 5:
            return PCI_ENTRY_REGEX.match(self.to_string(":")) is not None

        # Return False for anything else
        return False

    def get_biopython_key(self, full_id=False):
        """Represent the entry as a key to select chains or compounds from Biopython Entity objects.

        Parameters
        full_id : bool
            If True, return the full id of a chain or ligand.
            For chains, it consists of a tuple containing the PDB and the chain id.
            For ligands, it consists of a tuple containing the PDB, the chain, and the ligand id.
            The default value is False.

        Returns
        -------
         : str or tuple
            Return str if the entry represents a chain and if ``full_id`` is False.
            Otherwise, return a tuple.

        Examples
        --------

        >>> from luna.mol.entry import Entry
        >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, is_hetatm=True, sep=":")
        >>> print(e.get_biopython_key())
        ('H_X01', 300, ' ')

        """
        key = []
        if full_id:
            key = [self.pdb_id, 0, self.chain_id]

        if self.comp_name is not None and self.comp_num is not None:
            if self.comp_name == 'HOH' or self.comp_name == 'WAT':
                comp_id = ('W', self.comp_num, self.comp_icode)
            elif self.is_hetatm:
                comp_id = ('H_%s' % self.comp_name, self.comp_num, self.comp_icode)
            else:
                comp_id = (' ', self.comp_num, self.comp_icode)

            if full_id:
                key.append(comp_id)
                return tuple(key)
            return comp_id

        if full_id:
            return tuple(key)

        return self.chain_id

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.to_string(self.sep))


class ChainEntry(Entry):
    """Define a chain.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    sep : str
        A separator character to format the entry string. The default value is ':'.

    Raises
    ------
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------
    >>> from luna.mol.entry import ChainEntry
    >>> e = ChainEntry(pdb_id="3QL8", chain_id="A")
    >>> print(e)
    <ChainEntry: 3QL8:A>

    """

    def __init__(self, pdb_id, chain_id, sep=ENTRY_SEPARATOR, parser=None):
        super().__init__(pdb_id, chain_id, is_hetatm=False, sep=sep, parser=parser)

    @classmethod
    def from_string(cls, entry_str, sep=ENTRY_SEPARATOR):
        """Initialize from a string.

        Parameters
        ----------
        entry_str : str
            A string representing the entry. Example: '3QL8:A'.
        sep : str
            The separator character used in ``entry_str``. The default value is ':'.
            For example: if ``entry_str`` is set to '3QL8|A', then ``sep`` should be defined as '|'.

        Returns
        -------
         : `Entry`

        Raises
        ------
        IllegalArgumentError
            If the fields in ``entry_str`` do not match the format expected to define a chain.

        Examples
        --------
        >>> from luna.mol.entry import ChainEntry
        >>> e = ChainEntry.from_string("3QL8:A", sep=":")
        >>> print(e)
        <ChainEntry: 3QL8:A>

        """
        entries = entry_str.split(sep)
        if len(entries) == 2:
            return cls(*entries, sep=sep)
        else:
            raise IllegalArgumentError("The number of fields in the informed string '%s' is incorrect. A valid string must contain "
                                       "two obligatory fields: PDB and chain id." % entry_str)

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple (PDB id or filename, chain id)."""
        return (self.pdb_id, self.chain_id)


class MolEntry(Entry):
    """Define a compound from a PDB file, which can be a residue, nucleotide, or ligand.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    comp_name : str
        A 1 to 3-symbols compound name (residue name in the PDB format). Example: 'X01'.
    comp_num : int
        A valid 4-digits integer (residue sequence number in the PDB format). Example: 300 or -1.
    comp_icode : str, optional
        A 1-character compound insertion code (residue insertion code in the PDB format). Example: 'A'.
    sep : str
        A separator character to format the entry string. The default value is ':'.

    Raises
    ------
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------

    Compound entry: can be used to calculate interactions with a given compound (residue or nucleotide).

    >>> from luna.mol.entry import MolEntry
    >>> e = MolEntry(pdb_id="3QL8", chain_id="A", comp_name="HIS", comp_num=125, is_hetatm=False)
    >>> print(e)
    <MolEntry: 3QL8:A:HIS:125>

    Ligand entry: can be used to calculate interactions with a given ligand.

    >>> from luna.mol.entry import MolEntry
    >>> e = MolEntry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, is_hetatm=True)
    >>> print(e)
    <MolEntry: 3QL8:A:X01:300>

    """

    def __init__(self, pdb_id, chain_id, comp_name, comp_num,
                 comp_icode=None, sep=ENTRY_SEPARATOR, parser=None):

        super().__init__(pdb_id, chain_id, comp_name, comp_num, comp_icode,
                         is_hetatm=True, sep=sep, parser=parser)

    @classmethod
    def from_string(cls, entry_str, sep=ENTRY_SEPARATOR):
        entries = entry_str.split(sep)
        if len(entries) == 4:
            # Separate ligand number from insertion code.
            matched = REGEX_RESNUM_ICODE.match(entries[3])
            if matched:
                comp_num = matched.group(1)
                icode = None if matched.group(2) == "" else matched.group(2)
                entries = entries[0:3] + [comp_num, icode]
            else:
                raise IllegalArgumentError("The compound number and its insertion code (if applicable) '%s' is invalid. "
                                           "It must be an integer followed by one insertion code character when applicable." % entries[3])
            return cls(*entries, sep=sep)
        else:
            raise IllegalArgumentError("The number of fields in the informed string '%s' is incorrect. A valid compound entry must contain "
                                       "four obligatory fields: PDB, chain id, compound name, and compound number followed by its "
                                       "insertion code when applicable." % entry_str)


class MolFileEntry(Entry):

    """Define a ligand from a molecular file.
    This class should be used for docking and molecular dynamics campaigns where usually one has
    the protein structure in the PDB format and the ligand structure in a separate molecular file.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
    mol_id : str
        The ligand id in the molecular file.
    sep : str
        A separator character to format the entry string. The default value is ':'.

    Attributes
    ----------
    mol_id : str
        The ligand id.
    mol_file : str
        Pathname of the molecular file.
    mol_file_ext : str
        The molecular file format.
        If not provided, try to recover the molecular file extension directly from ``mol_file``.
    mol_obj_type : {'rdkit', 'openbabel'}
        Define which library (RDKit or Open Babel) to use to parse the molecular file.
    overwrite_mol_name : bool
        If True, substitute the ligand name in the parsed molecular object with ``mol_id``.
        Only works for single-molecule files (``is_multimol_file`` = False)
        as in these cases ``mol_id`` does not need to match the ligand name in the molecular file.
    is_multimol_file : bool
        If ``mol_file`` contains multiple molecules or not.
        If True, ``mol_id`` should match some ligand name in ``mol_file``.
    """

    def __init__(self, pdb_id, mol_id, sep=ENTRY_SEPARATOR):

        self.mol_id = mol_id

        #
        # Initialize empty properties.
        #
        self._mol_obj = None
        self.mol_file = None
        # TODO: Find a way to assume the Mol file type when not provided
        self.mol_file_ext = None
        self.mol_obj_type = None
        self.overwrite_mol_name = None
        self.is_multimol_file = None

        super().__init__(pdb_id, DEFAULT_CHAIN_ID, "LIG", 9999,
                         is_hetatm=True, sep=sep)

    @classmethod
    def from_mol_obj(cls, pdb_id, mol_id, mol_obj, sep=ENTRY_SEPARATOR):
        """Initialize from an already loaded molecular object.

        This function is useful in cases where a molecular object is parsed and
        pre-processed using a different protocol defined by the user.

        Parameters
        ----------
        pdb_id : str
            A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
        mol_id : str
            The ligand id.
            As the molecular object is already provided, the ligand id does not need to match the ligand name in the molecular object.
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
            The molecular object.
        sep : str
            A separator character to format the entry string. The default value is ':'.

        Returns
        -------
         : `MolFileEntry`

        Raises
        ------
        MoleculeObjectTypeError
            If the molecular object is not an instance
            of :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`.
        IllegalArgumentError
            If ``entity`` is not a valid Biopython object.

        Examples
        --------

        In this example, we will initialize a new MolFileEntry with the ligand 'ZINC000007786517' and
        the structure located in a PDB file of name 'D4', which is the structure used for docking the molecule.

        First, let's parse the molecular file.

        >>> from luna.wrappers.rdkit import read_mol_from_file
        >>> mol_obj = read_mol_from_file("tutorial/inputs/ZINC000007786517.mol", mol_format="mol")

        Now, we create the new MolFileEntry object as follows:

        >>> e = MolFileEntry.from_mol_obj("D4", "ZINC000007786517", mol_obj, sep=ENTRY_SEPARATOR)
        >>> print(e)
        <MolFileEntry: D4:ZINC000007786517>
        >>> print(e.mol_obj.to_smiles())
        Cc1cccc(NC(=O)C[N@@H+](C)C2CCCCC2)c1C
        """

        if mol_obj is not None:
            if isinstance(mol_obj, MolWrapper):
                mol_obj = mol_obj.unwrap()
            elif isinstance(mol_obj, PybelMol):
                mol_obj = mol_obj.OBMol

            if isinstance(mol_obj, RDMol):
                mol_obj_type = "rdkit"
            elif isinstance(mol_obj, OBMol):
                mol_obj_type = "openbabel"
            else:
                logger.exception("Objects of type '%s' are not currently "
                                 "accepted." % mol_obj.__class__)
                raise MoleculeObjectTypeError("Objects of type '%s' are not "
                                              "currently accepted."
                                              % mol_obj.__class__)
        else:
            if mol_obj_type not in ACCEPTED_MOL_OBJ_TYPES:
                raise IllegalArgumentError("Objects of type '%s' are not currently accepted. "
                                           "The available options are: %s." % (mol_obj_type, ", ".join(ACCEPTED_MOL_OBJ_TYPES)))

        entry = cls(pdb_id, mol_id, sep)
        entry.mol_obj = mol_obj
        entry.mol_obj_type = mol_obj_type

        # TODO: Find a way to assume the Mol file type when not provided
        # mol_file_ext

        return entry

    @classmethod
    def from_mol_file(cls, pdb_id, mol_id, mol_file, is_multimol_file,
                      mol_file_ext=None, mol_obj_type='rdkit',
                      autoload=False, overwrite_mol_name=False,
                      sep=ENTRY_SEPARATOR):
        """Initialize from a molecular file.

        Parameters
        ----------
        pdb_id : str
            A 4-symbols structure id from PDB or a local PDB filename. Example: '3QL8' or 'file1'.
        mol_id : str
            The ligand id in the molecular file.
        mol_file : str
            Pathname of the molecular file.
        is_multimol_file : bool
            If ``mol_file`` contains multiple molecules or not.
            If True, ``mol_id`` should match some ligand name in ``mol_file``.
        mol_file_ext : str, optional
            The molecular file format.
            If not provided, try to recover the molecular file extension directly from ``mol_file``.
        mol_obj_type : {'rdkit', 'openbabel'}
            If "rdkit", parse the converted molecule with RDKit and return an instance of :class:`rdkit.Chem.rdchem.Mol`.
            If "openbabel", parse the converted molecule with Open Babel and return an instance of
            :class:`openbabel.pybel.Molecule`. The default value is 'rdkit'.
        autoload : bool
            If True, parse the ligand from the molecular file during the entry initialization.
            Otherwise, only load the ligand when first used.
        overwrite_mol_name : bool
            If True, substitute the ligand name in the parsed molecular object with ``mol_id``.
            Only works for single-molecule files (``is_multimol_file`` = False)
            as in these cases ``mol_id`` does not need to match the ligand name in the molecular file.
        sep : str
            A separator character to format the entry string. The default value is ':'.

        Returns
        -------
         : `MolFileEntry`

        Raises
        ------
        FileNotFoundError
            If ``mol_file`` does not exist.
        IllegalArgumentError
            If ``mol_obj_type`` is not either 'rdkit' nor 'openbabel'.
        MoleculeObjectError
            If any errors occur while parsing the molecular file.
            Detailed information about the errors can be found in the logging outputs.
        MoleculeNotFoundError
            If the ligand ``mol_id`` was not found in the input file and ``is_multimol_file`` is True.

        Examples
        --------

        In this first example, we will read the ligand 'ZINC000007786517' from a single-molecule file.
        As we are working with a single-molecule file, ``mol_id`` can be any value you prefer.

        >>> from luna.mol.entry import MolFileEntry
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", mol_id="Ligand", mol_file="tutorial/inputs/ZINC000007786517.mol",
        ...                                mol_obj_type='rdkit', is_multimol_file=False)
        >>> print(e)
        D4:Ligand
        >>> print(e.mol_obj.to_smiles())
        Cc1cccc(NC(=O)C[N@@H+](C)C2CCCCC2)c1C

        Now, let's say we need to read the ligand 'ZINC000096459890' from a multi-molecular file and that we
        want to use Open Babel to parse the molecule.
        To do so, remember that it should exist a ligand with the name ``mol_id`` in ``mol_file``.
        Otherwise, it will raise the exception MoleculeNotFoundError.

        >>> from luna.mol.entry import MolFileEntry
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", mol_id="ZINC000096459890", mol_file="tutorial/inputs/ligands.mol2",
        ...                                mol_obj_type='openbabel', is_multimol_file=True)
        >>> print(e)
        <MolFileEntry: D4:ZINC000096459890>
        >>> print(e.mol_obj.to_smiles())
        O=C(OCCCN1C=CC=CC1=O)c1ccc2ccc(Cl)cc2n1

        Below, we show what happens if ``mol_id`` does not exist in ``mol_file``. Observe we set ``autoload`` to True
        to parse the molecule right away.

        >>> from luna.mol.entry import MolFileEntry
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", mol_id="Ligand", mol_file="tutorial/inputs/ligands.mol2",
        ...                                mol_obj_type='openbabel', is_multimol_file=True, autoload=True)
        luna.util.exceptions.MoleculeNotFoundError: "The ligand 'Ligand' was not found in the input file \
        or generated errors while parsing it with Open Babel."
        """

        entry = cls(pdb_id, mol_id, sep)

        entry.mol_file = mol_file
        entry.is_multimol_file = is_multimol_file
        entry.mol_file_ext = mol_file_ext or get_file_format(mol_file)
        entry.mol_obj_type = mol_obj_type
        entry.overwrite_mol_name = overwrite_mol_name

        if autoload:
            entry._load_mol_from_file()

        return entry

    @classmethod
    def from_file(cls, input_file, pdb_id=None, mol_file=None,
                  fields_sep=",", **kwargs):

        if not exists(input_file):
            raise OSError("File '%s' does not exist." % input_file)

        with open(input_file, "r") as IN:
            c = 1
            for row in IN:
                row = row.strip()
                if row == "":
                    continue

                args = row.split(fields_sep)

                if len(args) == 1:
                    if pdb_id is None or mol_file is None:
                        msg = (f"Only one field was found in line #{c}. "
                               "In these cases, 'pdb_id' and 'mol_file' "
                               "are mandatory.")
                        raise IllegalArgumentError(msg)

                    curr_pdb_id = pdb_id
                    curr_mol_id = args[0]
                    curr_mol_file = mol_file
                    is_multimol_file = True

                elif len(args) == 4:
                    (curr_pdb_id, curr_mol_id,
                        curr_mol_file, is_multimol_file) = args
                    is_multimol_file = ast.literal_eval(is_multimol_file)

                else:
                    msg = (f"Invalid number of arguments found in line #{c}. "
                           "You should provide either one or four arguments.")
                    raise IllegalArgumentError(msg)

                yield cls.from_mol_file(curr_pdb_id, curr_mol_id,
                                        curr_mol_file,
                                        is_multimol_file=is_multimol_file,
                                        **kwargs)
                c += 1

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple (PDB id or filename, ligand id)."""
        return (self.pdb_id, self.mol_id)

    @property
    def mol_obj(self):
        """:class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`: The molecule."""
        if self._mol_obj is None and self.mol_file is not None:
            self._load_mol_from_file()
        return self._mol_obj

    @mol_obj.setter
    def mol_obj(self, mol_obj):
        self._mol_obj = MolWrapper(mol_obj)

    def is_valid(self):
        """Check if the entry represents a valid protein-ligand complex.

        Returns
        -------
         : bool
        """
        return True

    def is_mol_obj_loaded(self):
        """Check if the molecular object has already been loaded.

        Returns
        -------
         : bool
        """
        return self._mol_obj is not None

    def _load_mol_from_file(self):
        logger.debug("It will try to load the molecule '%s'." % self.mol_id)

        if self.mol_file is None:
            raise IllegalArgumentError("It cannot load the molecule as no molecular file was provided.")

        available_formats = OB_FORMATS if self.mol_obj_type == "openbabel" else RDKIT_FORMATS
        tool = "Open Babel" if self.mol_obj_type == "openbabel" else "RDKit"
        if self.mol_file_ext not in available_formats:
            raise IllegalArgumentError("Extension '%s' informed or assumed from the filename is not a format "
                                       "recognized by %s." % (self.mol_file_ext, tool))

        if not exists(self.mol_file):
            raise FileNotFoundError("The file '%s' was not found." % self.mol_file)

        try:
            if self.mol_obj_type == "openbabel":
                mols = readfile(self.mol_file_ext, self.mol_file)
                # If it is a multimol file, then we need to loop over the molecules to find the target one.
                # Note that in this case, the ids must match.
                if self.is_multimol_file:
                    for ob_mol in mols:
                        if self.mol_id == get_filename(ob_mol.OBMol.GetTitle()):
                            self.mol_obj = ob_mol
                            break
                else:
                    self.mol_obj = mols.__next__()
            else:
                if self.mol_file_ext == "pdb":
                    self.mol_obj = read_mol_from_file(self.mol_file, mol_format=self.mol_file_ext, removeHs=False)
                else:
                    # If 'targets' is None, then the entire Mol file will be read.
                    targets = None
                    # If it is a multimol file than loop through it until the informed molecule (by its mol_id) is found.
                    if self.is_multimol_file:
                        targets = [self.mol_id]

                    for rdk_mol, mol_id in read_multimol_file(self.mol_file, mol_format=self.mol_file_ext, targets=targets, removeHs=False):
                        # It returns None if the molecule parsing generated errors.
                        self.mol_obj = rdk_mol
                        break
        except Exception as e:
            logger.exception(e)
            raise MoleculeObjectError("An error occurred while parsing the molecular file with %s and the molecule "
                                      "object for the entry '%s' could not be created. Check the logs for more information."
                                      % (tool, self.to_string()))

        if self._mol_obj is None:
            raise MoleculeNotFoundError("The ligand '%s' was not found in the input file or generated errors while parsing it with %s."
                                        % (self.mol_id, tool))
        else:
            if not self.mol_obj.has_name() or self.overwrite_mol_name:
                self.mol_obj.set_name(self.mol_id)

        logger.debug("Molecule '%s' was successfully loaded." % self.mol_id)

    def get_biopython_structure(self, entity=None, parser=None):
        """Transform the molecular object into a Biopython Entity object.

        If ``entity`` is provided, the molecular object is appended to it, i.e., this function
        can be used to join a ligand and the structure used during docking or molecular dynamics.

        By default, the ligand is added to a chain of id `z`.

        Parameters
        ----------
        entity : :class:`~luna.MyBio.PDB.Entity.Entity`, optional
            Append the molecular object to ``entity``.
            If not provided, a new :class:`~luna.MyBio.PDB.Entity.Entity` is created.
        parser : :class:`~luna.MyBio.PDB.PDBParser.PDBParser`, optional
            Define a PDB parser object. If not provided, the default parser will be used.

        Returns
        -------
         : :class:`~luna.MyBio.PDB.Entity.Entity`

        Raises
        ------
        IllegalArgumentError
            If ``entity`` is not a valid Biopython object.

        Examples
        --------

        In this example, we will demonstrate how to join a protein structure and a ligand docked against it.

        First, let's parse the PDB file.

        >>> from luna.MyBio.PDB.PDBParser import PDBParser
        >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
        >>> structure = pdb_parser.get_structure("Protein", "tutorial/inputs/D4.pdb")

        Observe that the list of chains in the parsed structure contains only one element.

        >>> print(structure[0].child_list)
        [<Chain id=A>]

        Now, we will read the ligand and append it to the existing protein structure.

        >>> from luna.mol.entry import MolFileEntry
        >>> e = MolFileEntry.from_mol_file("D4", "ZINC000007786517", "tutorial/inputs/ZINC000007786517.mol",
        ...                                mol_obj_type='rdkit', is_multimol_file=False)
        >>> joined_structure = e.get_biopython_structure(structure)

        Observe that now the list of chains contains chains 'A' and 'z', which is the default chain where
        ligands are added.

        >>> print(joined_structure[0].child_list)
        [<Chain id=A>, <Chain id=z>]

        If we loop over the residues in chain 'z', we will find our ligand.

        >>> for r in structure[0]["z"]:
        >>>     print(r)
        <Residue LIG het=H_LIG resseq=9999 icode= >
        """

        if parser is None:
            parser = PDBParser(PERMISSIVE=True, QUIET=True, FIX_EMPTY_CHAINS=True,
                               FIX_ATOM_NAME_CONFLICT=True, FIX_OBABEL_FLAGS=False)

        mol_file_ext = self.mol_file_ext
        if mol_file_ext is None and self.mol_file is not None:
            mol_file_ext = get_file_format(self.mol_file)

        if self.mol_obj_type == "openbabel":
            pdb_block = self.mol_obj.to_pdb_block()

            atm = self.mol_obj.unwrap().GetFirstAtom()
            residue_info = atm.GetResidue()

            # When the PDBParser finds an empty chain, it automatically replace it by 'z'.
            chain_id = residue_info.GetChain() if residue_info.GetChain().strip() != "" else self.chain_id
            comp_num = residue_info.GetNum()

            if residue_info.GetName() in WATER_NAMES:
                comp_name = "W"
            elif residue_info.IsHetAtom(atm):
                comp_name = "H_%s" % residue_info.GetName()
            else:
                comp_name = " "

            if mol_file_ext == "pdb":
                self.chain_id = chain_id
                self.comp_name = residue_info.GetName()
                self.comp_num = comp_num
                self.is_hetatm = residue_info.IsHetAtom(atm)
        else:
            pdb_block = self.mol_obj.to_pdb_block()

            if mol_file_ext == "pdb":
                residue_info = self.mol_obj.unwrap().GetAtoms()[0].GetPDBResidueInfo()
                # When the PDBParser finds an empty chain, it automatically replace it by 'z'.
                chain_id = residue_info.GetChainId() if residue_info.GetChainId().strip() != "" else self.chain_id
                comp_num = residue_info.GetResidueNumber()

                if residue_info.GetResidueName() in WATER_NAMES:
                    comp_name = "W"
                elif residue_info.GetIsHeteroAtom():
                    comp_name = "H_%s" % residue_info.GetResidueName()
                else:
                    comp_name = " "

                self.chain_id = chain_id
                self.comp_name = residue_info.GetResidueName()
                self.comp_num = comp_num
                self.is_hetatm = residue_info.GetIsHeteroAtom()
            else:
                # When the PDBParser finds an empty chain, it automatically replace it by 'z'.
                chain_id = self.chain_id
                comp_name = "H_UNL"
                comp_num = 1

        comp_structure = parser.get_structure_from_pdb_block(self.pdb_id, pdb_block)

        chain = comp_structure[0][chain_id]
        if self.chain_id != chain.id:
            chain.id = self.chain_id

        lig = chain[(comp_name, comp_num, " ")]

        # It only substitutes the ligand id if it is different from the id defined by the MolFileEntry object property.
        # This update will never happen when the ligand file is a PDB file as the ids are guaranteed to be equal.
        if lig.id != ("H_%s" % self.comp_name, self.comp_num, " "):
            lig.id = ("H_%s" % self.comp_name, self.comp_num, " ")

        lig.resname = self.comp_name

        if entity is not None:
            if isinstance(entity, Entity):
                structure = entity.get_parent_by_level('S')
                if self.chain_id not in structure[0].child_dict:
                    chain = chain.copy()
                    structure[0].add(chain)
                else:
                    lig = lig.copy()
                    # Update the ligand index according to the number of compounds already present in the chain.
                    lig.idx = len(structure[0][self.chain_id].child_list)
                    structure[0][self.chain_id].add(lig)
            else:
                raise IllegalArgumentError("The informed entity is not a valid Biopython object.")
        else:
            entity = comp_structure

        return entity

    def __repr__(self):
        return '<MolFileEntry: %s%s%s>' % (self.pdb_id, self.sep, self.mol_id)

    def __getstate__(self):
        if self._mol_obj is not None:
            self.mol_obj = MolWrapper(self.mol_obj)
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)


def recover_entries_from_entity(entity, get_small_molecules=True, get_chains=False,
                                ignore_artifacts=True, by_cluster=False, sep=ENTRY_SEPARATOR):
    """ Search for chains and small molecules in ``entity`` and return them as strings.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        An entity from where chains or small molecules will be recovered.
    get_small_molecules : bool
        If True, identify small molecules and return them as `MolEntry` objects. The default value is True.
    get_chains : bool
        If True, identify chains and return them as `ChainEntry` objects. The default value is False.
    ignore_artifacts : bool
        If True, ignore the following crystallography artifacts: ACE, ACT, BME, CSD, CSW, EDO, FMT, GOL, MSE,
        NAG, NO3, PO4, SGM, SO4, or TPO. The default value is True.
    by_cluster : bool
        If True, aggregate entries by cluster. Cluster ids are exclusive to :class:`~luna.MyBio.PDB.Residue.Residue` instances and are
        automatically set by :class:`~luna.MyBio.PDB.FTMapParser.FTMapParser`, a parser for FTMap results.
        By default, the cluster id of :class:`~luna.MyBio.PDB.Residue.Residue` instances are set to None, therefore,
        if the cluster id is not explicitly defined, all entries will be aggregated to the same key ``None``.
    sep : str
        A separator character to format the entry string. The default value is ':'.

    Returns
    ------
     : list or dict
        If ``by_cluster`` is set to False, a list of `ChainEntry` or `MolEntry` objects is returned.
        Otherwise, a dict is returned, in which keys are clusters and values are lists of `ChainEntry` or `MolEntry` objects.
        When no cluster information is available, all entries are aggregated in a key of value ``None``.
        Cluster ids are exclusive to :class:`~luna.MyBio.PDB.Residue.Residue` instances, therefore, `ChainEntry` objects
        are always placed in a key of value ``None``.

    Examples
    --------

    First, let's parse a PDB file.

    >>> from luna.MyBio.PDB.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein", "tutorial/inputs/3QQK.pdb")

    Now, we can recover entries from the parsed PDB file:

    >>> from luna.mol.entry import recover_entries_from_entity
    >>> entries = recover_entries_from_entity(structure, get_chains=True)
    >>> for e in entries:
    >>>     print(e)
    <MolEntry: Protein:A:X02:497>
    <ChainEntry: Protein:A>

    """

    clusters = defaultdict(list)
    entries = []

    if entity.level == "S":
        if get_small_molecules:
            residues = entity[0].get_residues()
        if get_chains:
            chains = entity[0].get_chains()

    elif entity.level == "M":
        if get_small_molecules:
            residues = entity.get_residues()
        if get_chains:
            chains = entity.get_chains()
    else:
        if get_small_molecules:
            # If the entity is already a Chain, get_parent_by_level() returns the same object.
            # But, if the entity is a Residue or Atom, it will return its corresponding chain parent.
            residues = entity.get_parent_by_level("C").get_residues()
        if get_chains:
            chains = entity.get_parent_by_level("M").get_chains()

    if get_small_molecules:
        pdb_id = entity.get_parent_by_level("S").id
        for res in residues:
            if res.is_hetatm():
                if ignore_artifacts and res.resname in ARTIFACTS_LIST:
                    continue

                comp_num_and_icode = ""
                if isinstance(res.id[1], int):
                    comp_num_and_icode = str(res.id[1])
                comp_num_and_icode += str(res.id[2]) if res.id[2].strip() else ""

                entry = MolEntry.from_string(sep.join([pdb_id, res.parent.id,
                                                       res.resname, comp_num_and_icode]))
                if by_cluster:
                    clusters[res.cluster_id].append(entry)
                else:
                    entries.append(entry)

    if get_chains:
        pdb_id = entity.get_parent_by_level("S").id
        for chain in chains:
            entry = ChainEntry.from_string(sep.join([pdb_id, chain.id]))
            if by_cluster:
                clusters[None].append(entry)
            else:
                entries.append(entry)

    if by_cluster:
        return clusters
    return entries



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.


###################################################################

# Modifications included by Alexandre Fassio (alexandrefassio@dcc.ufmg.br).
# Date: 19/02/2018.

# 1) The module __lt__ was overwritten to allow sorting in Python 3
# 2) Inherit inhouse modifications. Package: MyBio.

# Each line or block with modifications contain a MODBY tag.

###################################################################


"""Model class, used in Structure objects."""


# MODBY: Alexandre Fassio
# Inherit inhouse modifications. Package: MyBio.
from luna.MyBio.PDB.Entity import Entity


class Model(Entity):
    """The object representing a model in a structure. In a structure
    derived from an X-ray crystallography experiment, only a single
    model will be present (with some exceptions). NMR structures
    normally contain many different models.
    """

    def __init__(self, id, serial_num=None):
        """Initialize.

        Arguments:
         - id - int
         - serial_num - int

        """
        self.level = "M"
        if serial_num is None:
            self.serial_num = id
        else:
            self.serial_num = serial_num

        Entity.__init__(self, id)

    # Special methods

    # MODBY: Alexandre Fassio.
    # __lt__ method overwritten.
    def __lt__(self, r2):
        return self.id < r2.id

    # Private methods

    def _sort(self, c1, c2):
        """Sort the Chains instances in the Model instance.

        Chain instances are sorted alphabetically according to
        their chain id. Blank chains come last, as they often consist
        of waters.

        Arguments:
         - c1, c2 - Chain objects

        """
        id1 = c1.get_id()
        id2 = c2.get_id()
        # make sure blank chains come last (often waters)
        if id1 == " " and not id2 == " ":
            return 1
        elif id2 == " " and not id1 == " ":
            return -1
        return cmp(id1, id2)

    # Special methods

    def __repr__(self):
        return "<Model id=%s>" % self.get_id()

    # Public

    def get_chains(self):
        for c in self:
            yield c

    def get_residues(self):
        for c in self.get_chains():
            for r in c:
                yield r

    def get_atoms(self):
        for r in self.get_residues():
            for a in r:
                yield a
